import logging
import os
import pickle
import warnings
from typing import Sequence

import h5py
import pandas as pd

from dosma.utils.logger import setup_logger

__all__ = ["mkdirs", "save_pik", "load_pik", "save_h5", "load_h5", "save_tables"]


def mkdirs(dir_path: str):
    """Make directory is directory does not exist.

    Args:
        dir_path (str): Directory path to make.

    Returns:
        str: path to directory
    """
    if not os.path.isdir(dir_path):
        os.makedirs(dir_path)

    return dir_path


def save_pik(file_path: str, data):
    """Save data using `pickle`.

    Pickle is not be stable across Python 2/3.

    Args:
        file_path (str): File path to save to.
        data (Any): Data to serialize.
    """
    mkdirs(os.path.dirname(file_path))
    with open(file_path, "wb") as f:
        pickle.dump(data, f)


def load_pik(file_path: str):
    """Load data using `pickle`.

    Should be used with :any:`save_pik`.

    Pickle is not be stable across Python 2/3.

    Args:
        file_path (str): File path to load from.

    Returns:
        Any: Loaded data.

    Raises:
        FileNotFoundError: If `file_path` does not exist.
    """
    if not os.path.isfile(file_path):
        raise FileNotFoundError("{} does not exist".format(file_path))

    with open(file_path, "rb") as f:
        return pickle.load(f)


def save_h5(file_path: str, data_dict: dict):
    """Save data in H5DF format.

    Args:
        file_path (str): File path to save to.
        data_dict (dict): Dictionary of data to store. Dictionary can only have depth of 1.
    """
    mkdirs(os.path.dirname(file_path))
    with h5py.File(file_path, "w") as f:
        for key in data_dict.keys():
            f.create_dataset(key, data=data_dict[key])


def load_h5(file_path):
    """Load data in H5DF format.

    Args:
        file_path (str): File path to save to.

    Returns:
        dict: Loaded data.

    Raises:
        FileNotFoundError: If `file_path` does not exist.
    """
    if not os.path.isfile(file_path):
        raise FileNotFoundError("{} does not exist".format(file_path))

    data = {}
    with h5py.File(file_path, "r") as f:
        for key in f.keys():
            data[key] = f.get(key).value

    return data


def save_tables(
    file_path: str, data_frames: Sequence[pd.DataFrame], sheet_names: Sequence[str] = None
):
    """Save data in excel tables.

    Args:
        file_path (str): File path to excel file.
        data_frames (Sequence[pd.DataFrame]): Tables to store to excel file.
            One table stored per sheet.
        sheet_names (:obj:`Sequence[str]`, optional): Sheet names for each data frame.
    """
    mkdirs(os.path.dirname(file_path))
    writer = pd.ExcelWriter(file_path)

    if sheet_names is None:
        sheet_names = []
        for i in range(len(data_frames)):
            sheet_names.append("Sheet%d" % (i + 1))

    if len(data_frames) != len(sheet_names):
        raise ValueError("Number of data_frames and sheet_frames should be the same")

    for i in range(len(data_frames)):
        df = data_frames[i]
        df.to_excel(writer, sheet_names[i], index=False)

    writer.save()


def init_logger(log_file: str, debug: bool = False):  # pragma: no cover
    """Initialize logger.

    Args:
        log_file (str): File path to log file.
        debug (:obj:`bool`, optional): If ``True``, debug mode will be enabled for the logger.
            This means debug statements will also be written to the log file.
            Defaults to ``False``.
    """
    warnings.warn(
        "init_logger is deprecated since v0.0.14 and will no longer be "
        "supported in v0.13. Use `dosma.setup_logger` instead.",
        DeprecationWarning,
    )

    level = logging.DEBUG if debug else logging.INFO
    setup_logger(log_file, level=level)



"""The core module contains functions and classes for medical image analysis.
"""
from dosma.utils.logger import setup_logger  # noqa

from dosma import core as _core  # noqa: E402

from dosma.core import *  # noqa
from dosma.defaults import preferences  # noqa
from dosma.utils.collect_env import collect_env_info  # noqa
from dosma.utils.env import debug  # noqa

from dosma.core.med_volume import MedicalVolume  # noqa: F401
from dosma.core.io.format_io_utils import read, write  # noqa: F401
from dosma.core.io.format_io import ImageDataFormat  # noqa: F401
from dosma.core.io.dicom_io import DicomReader, DicomWriter  # noqa: F401
from dosma.core.io.nifti_io import NiftiReader, NiftiWriter  # noqa: F401
from dosma.core.device import Device, get_device, to_device  # noqa: F401
from dosma.core.orientation import to_affine  # noqa: F401
from dosma.core.registration import (  # noqa: F401
    register,
    apply_warp,
    symlink_elastix,
    unlink_elastix,
)
from dosma.core.fitting import (  # noqa: F401
    CurveFitter,
    PolyFitter,
    MonoExponentialFit,
    curve_fit,
    polyfit,
)

import dosma.core.numpy_routines as numpy_routines  # noqa: F401


__all__ = []
__all__.extend(_core.__all__)

setup_logger()

# This line will be programatically read/write by setup.py.
# Leave them at the bottom of this file and don't touch them.
__version__ = "0.1.1"



import os
from abc import ABC
from collections import defaultdict
from enum import Enum
from typing import Callable, Dict, Tuple, Union

import numpy as np
import pandas as pd

from dosma.core.io import format_io_utils as fio_utils
from dosma.core.io.format_io import ImageDataFormat
from dosma.core.med_volume import MedicalVolume
from dosma.defaults import preferences

__all__ = ["QuantitativeValueType", "QuantitativeValue", "T1Rho", "T2", "T2Star"]


class QuantitativeValueType(Enum):
    """Enum of types of quantitative values that can be analyzed.

    For more information on quantitative parameters, see :obj:`QuantitativeValue`.
    """

    T1_RHO = 1
    T2 = 2
    T2_STAR = 3


class QuantitativeValue(ABC):
    """This class handles tracking volumes associated with different quantitative values.

    Quantitative MRI characterizes the relaxation profile of different regions in the volume.
    This profile is determined by the composition of the object and has been shown to be
    informative for early detection of pathology.

    In practice, many of these quantitative relaxation parameters
    (:math:`T_1`, :math:`T_2`, :math:`T_2^*`, etc.) are computed per voxel by fitting
    to the exponential decay/recovery curves or by some analytical method. This results
    in a volume where different voxels have different relaxation parameters. These volumes
    are referred to as *volumetric quantitative maps* or *quantitative maps*.

    These fitted/computed relaxation parameters are called quantitative values.

    Args:
        volumetric_map (MedicalVolume, optional): Volumetric quantitative map.

    Attributes:
        volumetric_map (MedicalVolume): Volumetric quantitative map.
        additional_volumes (Dict[str, MedicalVolume]): Additional volumes associated
            with quantitative value. These are typically volumes associated with the
            goodness of fit of the value. For example, a volume could be pixel-wise r-squared,
            or error bounds, etc.

    Note:
        This class is in alpha - signature and functionality changes are very likely.
    """

    ID = 0
    NAME = ""

    def __init__(self, volumetric_map: MedicalVolume = None):
        assert self.ID > 0, "Attribute `ID` not initialized for {}".format(type(self))
        assert self.NAME != "", "Attribute `NAME` not initialized for {}".format(type(self))

        # Main 3D quantitative value map (MedicalVolume)
        if volumetric_map is not None and not isinstance(volumetric_map, MedicalVolume):
            raise TypeError("`volumetric_map` must be of type MedicalVolume")

        self.volumetric_map = volumetric_map

        # Sometimes there are additional volumes that we want to keep track of
        # i.e. r2 values, goodness of fit, etc
        # Store these as string_name: MedicalVolume (i.e. "r2: MedicalVolume)
        # To add values to this field, use the add_additional_volume method below
        # these results will not be loaded
        self.additional_volumes = {}

    def save_data(
        self, dir_path: str, data_format: ImageDataFormat = preferences.image_data_format
    ):
        """Save data to disk.

        Data will be stored in folder '`dir_path`/`self.NAME`'.

        Args:
            dir_path (str): Directory path.
            data_format (:obj:`ImageDataFormat`, optional): Data format to save medical volumes.
                Defaults to ``preferences.image_data_format``.
        """
        if data_format != ImageDataFormat.nifti:
            import warnings

            warnings.warn(
                "Due to bit depth issues, only nifti format is supported for quantitative values. "
                "Writing as nifti file..."
            )
            data_format = ImageDataFormat.nifti

        if self.volumetric_map is not None:
            filepath = os.path.join(dir_path, self.NAME, "{}.nii.gz".format(self.NAME))
            # filepath = fio_utils.convert_format_filename(filepath, data_format)
            self.volumetric_map.save_volume(filepath, data_format=data_format)

        for volume_name in self.additional_volumes.keys():
            add_vol_filepath = os.path.join(
                dir_path, self.NAME, "{}-{}.nii.gz".format(self.NAME, volume_name)
            )
            # add_vol_filepath = fio_utils.convert_format_filename(add_vol_filepath, data_format)
            self.additional_volumes[volume_name].save_volume(
                add_vol_filepath, data_format=data_format
            )

    def load_data(self, dir_path):
        """Load data from disk.

        Data will be loaded from folder '`dir_path`/`self.NAME`'.

        Currently, additional volumes are not reloaded.

        Args:
            dir_path (str): Directory path.
        """
        file_path = os.path.join(dir_path, self.NAME, "{}.nii.gz".format(self.NAME))
        qv_volume = fio_utils.generic_load(file_path, expected_num_volumes=1)

        self.volumetric_map = qv_volume

    def add_additional_volume(self, name: str, volume: MedicalVolume):
        """Add volume that corresponds to quantitative value.

        Additional volumes are typically volumes associated with the goodness of fit of the value.
        For example, a volume could be r-squared values per voxel, or error bounds, etc.

        This should not be the volumetric quantitative map.
        To update that map, see ``self.volumetric_map``.

        Args:
            name (str): Name of additional volume.
            volume (MedicalVolume):
        """
        if not isinstance(volume, MedicalVolume):
            raise TypeError("`volumes` must be of type MedicalVolume")
        self.additional_volumes[name] = volume

    def to_metrics(
        self,
        mask: MedicalVolume = None,
        labels: Dict[int, str] = None,
        bounds: Tuple[float, float] = None,
        closed: str = "right",
        fns: Dict[str, Callable] = None,
    ) -> pd.DataFrame:
        """Compute scalar metrics for quantitative values.

        Metrics include mean, median, standard deviation, and number of voxels.
        Valid voxels are defined as finite valued voxels within the interval
        `bounds` (if specified).

        Args:
            mask (MedicalVolume, optional): Label mask. Labels should be unsigned ints (uint).
                Metrics will be computed for each non-zero label(s).
                If `labels` specified, metrics computed only for keys in `labels` dictionary.
                If not specified, metrics will be calculated over all valid voxels.
            labels (Dict[int, str], optional): Mapping from label to label name.
                If specified, only labels in this argument will be computed.
            bounds (float or Tuple[float, float], optional): The left/right bounds
                for computing metrics. By default, the bounds are the
                open-interval `(-inf, inf)`.
            closed (str, optional): If `bounds` specified, whether the bounds are closed
                on the left-side, right-side, both or neither.
                One of {'right', 'left', 'both', 'neither'}.

        Returns:
            metrics (pd.DataFrame): Metrics for quantitative value. Columns include:
                * "Region" (str): The label name.
                * "Mean" (float): Average quantitative value in a region.
                * "Median" (float): Median quantitative value in a region.
                * "Std" (float): Standard deviation of quantitative value in a region.
                * "# Voxels" (int): The number of valid voxels in the region.
        """
        volume = self.volumetric_map.volume
        valid_mask = np.isfinite(volume)
        if bounds:
            assert len(bounds) == 2, len(bounds)  # Expected (left,right) bound
            lb, ub = bounds[0], bounds[1]
            assert lb <= ub, f"lower:{lb}, upper: {ub}"  # Expected left bound <= right bound
            assert closed in ("right", "left", "both", "neither"), closed
            lb_mask = volume >= lb if closed in ("left", "both") else volume > lb
            ub_mask = volume <= ub if closed in ("right", "both") else volume < ub
            valid_mask &= lb_mask & ub_mask

        if mask is not None:
            mask = mask.reformat(self.volumetric_map.orientation)
            mask = mask.volume

            if labels is None:
                unique_vals = [x for x in np.unique(mask) if x > 0]
                labels = {int(i): f"label_{int(i)}" for i in unique_vals}
            labels.update({-1: "total"})
        else:
            labels = {-2: "total"}

        if mask is not None:
            # Do not modify original array.
            mask = mask.copy()
            mask[~valid_mask] = 0

        if fns is None:
            fns = {}

        metrics = defaultdict(list)
        for label, name in labels.items():
            if label == -2:
                qv_region_vals = volume[valid_mask]  # Entire volume.
            elif label == -1:
                qv_region_vals = volume[mask > 0]  # noqa
            else:
                qv_region_vals = volume[mask == label]
            num_voxels = np.prod(qv_region_vals.shape)

            metrics["Category"].append(name)
            metrics["Mean"].append(np.nanmean(qv_region_vals))
            metrics["Std"].append(np.nanstd(qv_region_vals))
            metrics["Median"].append(np.nanmedian(qv_region_vals))
            metrics["# Voxels"].append(num_voxels)
            for name, fn in fns.items():
                metrics[name].append(fn(qv_region_vals))

        return pd.DataFrame(metrics)

    def to(self, device):
        """Moves volumes of this scan onto the appropriate device.

        This is an inplace operation. A new ScanSequence will not
        be returned.

        Args:
            device: The device to move to.

        Returns:
            self
        """
        self.volumetric_map = self.volumetric_map.to(device)
        self.additional_volumes = {k: v.to(device) for k, v in self.additional_volumes.items()}

        return self

    @staticmethod
    def get_qv(qv_id: Union[int, str]):
        """Find QuantitativeValue enum using id or name.

        Args:
            qv_id (:obj:`int` or :obj:`str`): Either quantitative value enum number
                or name in lower case.

        Returns:
            QuantitativeValue: Quantitative value corresponding to id.

        Raises:
            ValueError: If no QuantitativeValue corresponding to `qv_id` found.
        """
        for qv in [T1Rho(), T2(), T2Star()]:
            if qv.NAME.lower() == qv_id or qv.NAME == qv_id or qv.ID == qv_id:
                return qv

        raise ValueError("Quantitative Value with name or id {} not found".format(qv_id))

    @staticmethod
    def save_qvs(dir_path: str, qvs):
        """Save all quantitative values from directory.

        Args:
            dir_path (str): Directory path.
            qvs (Sequence[QuantitativeValue]): Quantitative values to save.
        """
        for qv in qvs:
            if not isinstance(qv, QuantitativeValue):
                raise TypeError("All members of `qvs` must be instances of QuantitativeValue")
            qv.save_data(dir_path)

    @staticmethod
    def load_qvs(dir_path: str):
        """Load all quantitative values from directory.

        Args:
            dir_path (str): Directory path.

        Returns:
            list[QuantitativeValue]: Quantitative value wrappers.
        """
        qvs = []
        for qv in [T1Rho(), T2(), T2Star()]:
            possible_qv_filepath = os.path.join(dir_path, qv.NAME, "{}.nii.gz".format(qv.NAME))
            if os.path.isfile(possible_qv_filepath):
                qv.load_data(dir_path)
                qvs.append(qv)

        return qvs

    @property
    def qv_type(self) -> QuantitativeValueType:
        """QuantitativeValueType: quantitative value type."""
        raise NotImplementedError(f"Quantitative value type not implemented for {type(self)}")


class T1Rho(QuantitativeValue):
    """T1Rho MRI parameter."""

    ID = 1
    NAME = "t1_rho"

    @property
    def qv_type(self):
        return QuantitativeValueType.T1_RHO


class T2(QuantitativeValue):
    """T2 MRI parameter."""

    ID = 2
    NAME = "t2"

    @property
    def qv_type(self):
        return QuantitativeValueType.T2


class T2Star(QuantitativeValue):
    """T2Star MRI parameter."""

    ID = 3
    NAME = "t2_star"

    @property
    def qv_type(self):
        return QuantitativeValueType.T2_STAR



# bool --> booltype, default, checkbox
# filepath (str) --> stringtype, default, load file
# str --> stringtype, default, textbox
# float/int -->

import inspect
import tkinter as tk
from tkinter import ttk

from dosma.cli import BASIC_TYPES
from dosma.gui.gui_utils.filedialog_reader import FileDialogReader


class TextWithVar(tk.Text):
    """A text widget that accepts a 'textvariable' option"""

    def __init__(self, parent, *args, **kwargs):
        try:
            self._textvariable = kwargs.pop("textvariable")
        except KeyError:
            self._textvariable = None

        tk.Text.__init__(self, parent, *args, **kwargs)

        # if the variable has data in it, use it to initialize
        # the widget
        if self._textvariable is not None:
            self.insert("1.0", self._textvariable.get())

        # this defines an internal proxy which generates a
        # virtual event whenever text is inserted or deleted
        self.tk.eval(
            """
            proc widget_proxy {widget widget_command args} {

                # call the real tk widget command with the real args
                set result [uplevel [linsert $args 0 $widget_command]]

                # if the contents changed, generate an event we can bind to
                if {([lindex $args 0] in {insert replace delete})} {
                    event generate $widget <<Change>> -when tail
                }
                # return the result from the real widget command
                return $result
            }
            """
        )

        # this replaces the underlying widget with the proxy
        self.tk.eval(
            """
            rename {widget} _{widget}
            interp alias {{}} ::{widget} {{}} widget_proxy {widget} _{widget}
        """.format(
                widget=str(self)
            )
        )

        # set up a binding to update the variable whenever
        # the widget changes
        self.bind("<<Change>>", self._on_widget_change)

        # set up a trace to update the text widget when the
        # variable changes
        if self._textvariable is not None:
            self._textvariable.trace("wu", self._on_var_change)

    def _on_var_change(self, *args):
        """Change the text widget when the associated textvariable changes"""

        # only change the widget if something actually
        # changed, otherwise we'll get into an endless
        # loop
        text_current = self.get("1.0", "end-1c")
        var_current = self._textvariable.get()
        if text_current != var_current:
            self.delete("1.0", "end")
            self.insert("1.0", var_current)

    def _on_widget_change(self, event=None):
        """Change the variable when the widget changes"""
        if self._textvariable is not None:
            self._textvariable.set(self.get("1.0", "end-1c"))


class Filepath(str):
    pass


TYPE_CAST = {bool: tk.BooleanVar, str: tk.StringVar, int: tk.IntVar, float: tk.DoubleVar}


def contains_filepath_keywords(param_name: str):
    fp_keywords = ["dir", "path", "directory", "file"]
    for k in fp_keywords:
        if k in param_name:
            return True

    return False


def convert_base_type_to_gui(param_name, param_type, param_default, root, **kwargs):
    balloon = None
    param_help = ""
    if "balloon" in kwargs:
        balloon = kwargs.get("balloon")
    if "param_help" in kwargs:
        param_help = kwargs.get("param_help")

    assert param_type in BASIC_TYPES, "type %s not in BASIC_TYPES" % param_type

    # add default value to param help
    has_default = param_default is not inspect._empty and param_default is not None

    type_var = TYPE_CAST[param_type]()
    if has_default:
        type_var.set(param_default)

    is_filepath = (
        (param_type is str) and (not has_default) and contains_filepath_keywords(param_name)
    )
    hbox = None
    if is_filepath:
        hbox = format_filepath_gui(root, param_name, type_var)
    elif param_type is bool:
        hbox = format_bool_gui(root, param_name, type_var)
    elif param_type is str:
        if "options" in kwargs:
            hbox = format_list_gui(root, param_name, type_var, **kwargs)
        else:
            hbox = format_str_gui(root, param_name, type_var)

    # TODO: Add suport for float and int values.
    if hbox:
        if balloon and param_help:
            balloon.bind(hbox, param_help)

    return type_var


def format_filepath_gui(root, label, type_var, **kwargs):
    hbox = tk.Frame(root)
    hbox.pack(side="top", anchor="nw")

    _label = tk.Label(hbox, text="%s: " % label)
    _label.pack(side="left", anchor="nw", padx=5)

    t = tk.Label(hbox, textvariable=type_var)
    t.pack(side="left", anchor="nw", padx=5)

    fd = FileDialogReader(type_var)
    fd_prompt = "Load %s" % label.lower()

    f_action = fd.get_filepath
    if "dir" in label.lower():
        f_action = fd.get_dirpath

    b = ttk.Button(root, text=fd_prompt, command=lambda: f_action(title=fd_prompt))

    b.pack(anchor="nw", pady=1)

    return hbox


def format_str_gui(root, label, type_var, **kwargs):
    hbox = tk.Frame(root)
    hbox.pack(side="top", anchor="nw")

    _label = tk.Label(hbox, text="%s: " % label)
    _label.pack(side="left", anchor="nw", padx=5)

    t = TextWithVar(hbox, textvariable=type_var)
    t.pack(side="left", anchor="nw", padx=5)

    return hbox


def format_bool_gui(root, label, type_var, **kwargs):
    hbox = tk.Frame(root)
    hbox.pack(side="top", anchor="nw")

    _label = tk.Label(hbox, text="%s: " % label)
    _label.pack(side="left", anchor="nw", padx=5)

    t = tk.Checkbutton(hbox, variable=type_var)
    t.pack(side="left", anchor="nw", padx=5)

    return hbox


def format_list_gui(root, label, type_var, **kwargs):
    options = kwargs.get("options")

    hbox = tk.Frame(root)
    hbox.pack(side="top", anchor="nw")

    _label = tk.Label(hbox, text="%s: " % label)
    _label.pack(side="left", anchor="nw", padx=5)

    t = tk.OptionMenu(hbox, type_var, *options)
    t.pack(side="left", anchor="nw", padx=5)

    return hbox



"""Analysis for meniscus.

Attributes:
    BOUNDS (dict): Upper bounds for quantitative values.
"""

import itertools
import os
import warnings

import numpy as np
import pandas as pd
import scipy.ndimage as sni

from dosma.core.device import get_array_module
from dosma.core.med_volume import MedicalVolume
from dosma.core.quant_vals import T2, QuantitativeValueType
from dosma.defaults import preferences
from dosma.tissues.tissue import Tissue, largest_cc
from dosma.utils import io_utils

import matplotlib.pyplot as plt

# milliseconds
BOUNDS = {
    QuantitativeValueType.T2: 60.0,
    QuantitativeValueType.T1_RHO: 100.0,
    QuantitativeValueType.T2_STAR: 50.0,
}

__all__ = ["Meniscus"]


class Meniscus(Tissue):
    """Handles analysis and visualization for meniscus.

    This class extends functionality from `Tissue`.

    For visualization, the meniscus is unrolled across the axial plane.
    """

    ID = 2
    STR_ID = "men"
    FULL_NAME = "meniscus"

    # Expected quantitative values
    T1_EXPECTED = 1000  # milliseconds

    # Coronal Keys
    _ANTERIOR_KEY = 0
    _POSTERIOR_KEY = 1
    _CORONAL_KEYS = [_ANTERIOR_KEY, _POSTERIOR_KEY]

    # Saggital Keys
    _MEDIAL_KEY = 0
    _LATERAL_KEY = 1
    _SAGGITAL_KEYS = [_MEDIAL_KEY, _LATERAL_KEY]

    # Axial Keys
    _SUPERIOR_KEY = 0
    _INFERIOR_KEY = 1
    _TOTAL_AXIAL_KEY = -1

    def __init__(
        self, weights_dir: str = None, medial_to_lateral: bool = None, split_ml_only: bool = False
    ):
        super().__init__(weights_dir=weights_dir, medial_to_lateral=medial_to_lateral)

        self.split_ml_only = split_ml_only
        self.regions_mask = None

    def unroll_axial(self, quant_map: np.ndarray):
        """Unroll meniscus in axial direction.

        Args:
            quant_map (np.ndarray): Map to roll out.

        """
        mask = self.__mask__.volume

        assert (
            self.regions_mask is not None
        ), "region_mask not initialized. Should be initialized when mask is set"
        region_mask_sup_inf = self.regions_mask[..., 0]

        superior = (region_mask_sup_inf == self._SUPERIOR_KEY) * mask * quant_map
        superior[superior == 0] = np.nan
        superior = np.nanmean(superior, axis=0)

        inferior = (region_mask_sup_inf == self._INFERIOR_KEY) * mask * quant_map
        inferior[inferior == 0] = np.nan
        inferior = np.nanmean(inferior, axis=0)

        total = mask * quant_map
        total[total == 0] = np.nan
        total = np.nanmean(total, axis=0)

        return total, superior, inferior

    def split_regions(self, base_map):
        """Split meniscus into subregions.

        Center-of-mass (COM) is used to subdivide into
        anterior/posterior, superior/inferior, and medial/lateral regions.

        Note:
            The anterior/posterior and superior/inferior subdivision may causes issues
            with tilted mensici. This will be addressed in a later release. To avoid
            computing metrics on these regions, set ``self.split_ml_only=True``.
        """
        center_of_mass = sni.measurements.center_of_mass(base_map)  # zero indexed

        com_sup_inf = int(np.ceil(center_of_mass[0]))
        com_ant_post = int(np.ceil(center_of_mass[1]))
        com_med_lat = int(np.ceil(center_of_mass[2]))

        region_mask_sup_inf = np.zeros(base_map.shape)
        region_mask_sup_inf[:com_sup_inf, :, :] = self._SUPERIOR_KEY
        region_mask_sup_inf[com_sup_inf:, :, :] = self._INFERIOR_KEY

        region_mask_ant_post = np.zeros(base_map.shape)
        region_mask_ant_post[:, :com_ant_post, :] = self._ANTERIOR_KEY
        region_mask_ant_post[:, com_ant_post:, :] = self._POSTERIOR_KEY

        region_mask_med_lat = np.zeros(base_map.shape)
        region_mask_med_lat[:, :, :com_med_lat] = (
            self._MEDIAL_KEY if self.medial_to_lateral else self._LATERAL_KEY
        )
        region_mask_med_lat[:, :, com_med_lat:] = (
            self._LATERAL_KEY if self.medial_to_lateral else self._MEDIAL_KEY
        )

        self.regions_mask = np.stack(
            [region_mask_sup_inf, region_mask_ant_post, region_mask_med_lat], axis=-1
        )

    def __calc_quant_vals__(self, quant_map: MedicalVolume, map_type: QuantitativeValueType):
        subject_pid = self.pid

        # Reformats the quantitative map to the appropriate orientation.
        super().__calc_quant_vals__(quant_map, map_type)

        assert (
            self.regions_mask is not None
        ), "region_mask not initialized. Should be initialized when mask is set"

        region_mask = self.regions_mask
        axial_region_mask = self.regions_mask[..., 0]
        coronal_region_mask = self.regions_mask[..., 1]
        sagittal_region_mask = self.regions_mask[..., 2]

        # Combine region mask into categorical mask.
        axial_categories = [
            (self._SUPERIOR_KEY, "superior"),
            (self._INFERIOR_KEY, "inferior"),
            (-1, "total"),
        ]
        coronal_categories = [
            (self._ANTERIOR_KEY, "anterior"),
            (self._POSTERIOR_KEY, "posterior"),
            (-1, "total"),
        ]
        sagittal_categories = [(self._MEDIAL_KEY, "medial"), (self._LATERAL_KEY, "lateral")]
        if self.split_ml_only:
            axial_categories = [x for x in axial_categories if x[0] == -1]
            coronal_categories = [x for x in coronal_categories if x[0] == -1]

        categorical_mask = np.zeros(region_mask.shape[:-1])
        base_mask = self.__mask__.A.astype(np.bool)
        labels = {}
        for idx, (
            (axial, axial_name),
            (coronal, coronal_name),
            (sagittal, sagittal_name),
        ) in enumerate(
            itertools.product(axial_categories, coronal_categories, sagittal_categories)
        ):
            label = idx + 1
            axial_map = np.asarray([True]) if axial == -1 else axial_region_mask == axial
            coronal_map = np.asarray([True]) if coronal == -1 else coronal_region_mask == coronal
            sagittal_map = sagittal_region_mask == sagittal
            categorical_mask[base_mask & axial_map & coronal_map & sagittal_map] = label
            labels[label] = f"{axial_name}-{coronal_name}-{sagittal_name}"

        # TODO: Change this to be any arbitrary quantitative value type.
        # Note, it does not matter what we wrap it in because the underlying operations
        # are not specific to the value type.
        t2 = T2(quant_map)
        categorical_mask = MedicalVolume(categorical_mask, affine=quant_map.affine)
        df = t2.to_metrics(categorical_mask, labels=labels, bounds=(0, np.inf), closed="neither")
        df.insert(0, "Subject", subject_pid)

        total, superior, inferior = self.unroll_axial(quant_map.volume)
        qv_name = map_type.name
        maps = [
            {
                "title": "%s superior" % qv_name,
                "data": superior,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_superior" % qv_name,
                "raw_data_filename": "%s_superior.data" % qv_name,
            },
            {
                "title": "%s inferior" % qv_name,
                "data": inferior,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_inferior" % qv_name,
                "raw_data_filename": "%s_inferior.data" % qv_name,
            },
            {
                "title": "%s total" % qv_name,
                "data": total,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_total" % qv_name,
                "raw_data_filename": "%s_total.data" % qv_name,
            },
        ]

        self.__store_quant_vals__(maps, df, map_type)

    def __calc_quant_vals_old__(self, quant_map, map_type):
        subject_pid = self.pid

        super().__calc_quant_vals__(quant_map, map_type)

        assert (
            self.regions_mask is not None
        ), "region_mask not initialized. Should be initialized when mask is set"

        quant_map_volume = quant_map.volume
        mask = self.__mask__.volume

        quant_map_volume = mask * quant_map_volume

        axial_region_mask = self.regions_mask[..., 0]
        sagittal_region_mask = self.regions_mask[..., 1]
        coronal_region_mask = self.regions_mask[..., 2]

        axial_names = ["superior", "inferior", "total"]
        coronal_names = ["medial", "lateral"]
        sagittal_names = ["anterior", "posterior"]

        pd_header = ["Subject", "Location", "Side", "Region", "Mean", "Std", "Median"]
        pd_list = []

        for axial in [self._SUPERIOR_KEY, self._INFERIOR_KEY, self._TOTAL_AXIAL_KEY]:
            if axial == self._TOTAL_AXIAL_KEY:
                axial_map = np.asarray(
                    axial_region_mask == self._SUPERIOR_KEY, dtype=np.float32
                ) + np.asarray(axial_region_mask == self._INFERIOR_KEY, dtype=np.float32)
                axial_map = np.asarray(axial_map, dtype=np.bool)
            else:
                axial_map = axial_region_mask == axial

            for coronal in [self._MEDIAL_KEY, self._LATERAL_KEY]:
                for sagittal in [self._ANTERIOR_KEY, self._POSTERIOR_KEY]:
                    curr_region_mask = (
                        quant_map_volume
                        * (coronal_region_mask == coronal)
                        * (sagittal_region_mask == sagittal)
                        * axial_map
                    )
                    curr_region_mask[curr_region_mask == 0] = np.nan
                    # discard all values that are 0
                    c_mean = np.nanmean(curr_region_mask)
                    c_std = np.nanstd(curr_region_mask)
                    c_median = np.nanmedian(curr_region_mask)

                    row_info = [
                        subject_pid,
                        axial_names[axial],
                        coronal_names[coronal],
                        sagittal_names[sagittal],
                        c_mean,
                        c_std,
                        c_median,
                    ]

                    pd_list.append(row_info)

        # Generate 2D unrolled matrix
        total, superior, inferior = self.unroll_axial(quant_map.volume)

        df = pd.DataFrame(pd_list, columns=pd_header)
        qv_name = map_type.name
        maps = [
            {
                "title": "%s superior" % qv_name,
                "data": superior,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_superior" % qv_name,
                "raw_data_filename": "%s_superior.data" % qv_name,
            },
            {
                "title": "%s inferior" % qv_name,
                "data": inferior,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_inferior" % qv_name,
                "raw_data_filename": "%s_inferior.data" % qv_name,
            },
            {
                "title": "%s total" % qv_name,
                "data": total,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_total" % qv_name,
                "raw_data_filename": "%s_total.data" % qv_name,
            },
        ]

        self.__store_quant_vals__(maps, df, map_type)

    def set_mask(self, mask: MedicalVolume, use_largest_ccs: bool = False, ml_only: bool = False):
        xp = get_array_module(mask.A)
        if use_largest_ccs:
            msk = xp.asarray(largest_cc(mask.A, num=2), dtype=xp.uint8)
        else:
            msk = xp.asarray(mask.A, dtype=xp.uint8)
        mask_copy = mask._partial_clone(volume=msk)
        super().set_mask(mask_copy)

        self.split_regions(self.__mask__.volume)

    def __save_quant_data__(self, dirpath):
        """Save quantitative data and 2D visualizations of meniscus

        Check which quantitative values (T2, T1rho, etc) are defined for meniscus and analyze these
            1. Save 2D total, superficial, and deep visualization maps
            2. Save {'medial', 'lateral'}, {'anterior', 'posterior'},
                {'superior', 'inferior', 'total'} data to excel file.

        Args:
            dirpath (str): Directory path to tissue data.
        """
        q_names = []
        dfs = []

        for quant_val in QuantitativeValueType:
            if quant_val.name not in self.quant_vals.keys():
                continue

            q_names.append(quant_val.name)
            q_val = self.quant_vals[quant_val.name]
            dfs.append(q_val[1])

            q_name_dirpath = io_utils.mkdirs(os.path.join(dirpath, quant_val.name.lower()))
            for q_map_data in q_val[0]:
                filepath = os.path.join(q_name_dirpath, q_map_data["filename"])
                xlabel = "Slice"
                ylabel = ""
                title = q_map_data["title"]
                data_map = q_map_data["data"]

                plt.clf()

                upper_bound = BOUNDS[quant_val]

                if preferences.visualization_use_vmax:
                    # Hard bounds - clipping
                    plt.imshow(data_map, cmap="jet", vmin=0.0, vmax=BOUNDS[quant_val])
                else:
                    # Try to use a soft bounds
                    if np.sum(data_map <= upper_bound) == 0:
                        plt.imshow(data_map, cmap="jet", vmin=0.0, vmax=BOUNDS[quant_val])
                    else:
                        warnings.warn(
                            "%s: Pixel value exceeded upper bound (%0.1f). Using normalized scale."
                            % (quant_val.name, upper_bound)
                        )
                        plt.imshow(data_map, cmap="jet")

                plt.xlabel(xlabel)
                plt.ylabel(ylabel)
                plt.title(title)
                clb = plt.colorbar()
                clb.ax.set_title("(ms)")
                plt.axis("tight")

                plt.savefig(filepath)

                # Save data
                raw_data_filepath = os.path.join(
                    q_name_dirpath, "raw_data", q_map_data["raw_data_filename"]
                )
                io_utils.save_pik(raw_data_filepath, data_map)

        if len(dfs) > 0:
            io_utils.save_tables(os.path.join(dirpath, "data.xlsx"), dfs, q_names)



"""Abstract classes interface implementation of scan.

This module defines the abstract interfaces that can be extended by concrete scan types.

Different scan types produce images with different qualities
(different quantitative parameters, resolutions, etc.).
These scan types have different actions (or processing) associated with them.

A scan can have multiple volumes if multiple phases or echo times are used to image in the scan.

Scans that have high resolution & signal-to-noise ratio (SNR) are typically used for
segmentation. Because of this property, they are often referred to as *targets*.
This allows them to serve as a good template that other lower SNR/resolution scans
can be registered to.
"""

import logging
import os
import re
from abc import abstractmethod
from time import localtime, strftime
from typing import Any, Sequence, Union

import numpy as np
import scipy.ndimage as sni
from natsort import natsorted

from dosma.core.io import format_io_utils as fio_utils
from dosma.core.io.nifti_io import NiftiReader
from dosma.core.med_volume import MedicalVolume
from dosma.defaults import preferences
from dosma.scan_sequences.scan_io import ScanIOMixin
from dosma.tissues.tissue import Tissue
from dosma.utils import env, io_utils

__all__ = ["ScanSequence"]

_logger = logging.getLogger(__name__)


class ScanSequence(ScanIOMixin):
    """The class for scan sequences and corresponding analysis.

    This is the base class for scan-specific analysis. All classes implementing scan-specific
    image processing/analysis methods should inherit from this class.

    Args:
        volumes (MedicalVolume(s)): The medical image(s) that this scan is composed of.
            Note these are typically the different 2D/3D image(s) . For example, for a
            multi-echo MRI scan, each volume in ``volumes`` would correspond to the spatial
            volume acquired at different echo times.

    Attributes:
        volumes (MedicalVolume(s)): See ``volume`` in Args.
        temp_path (str): The directory path where temporary results
            are written. This will be under ``dosma.file_constants.TEMP_FOLDER_PATH``.

    Raises:
        NotADirectoryError: If `dicom_path` is not a valid directory.
        ValueError: If dicoms do not correspond to the expected sequence.
    """

    NAME = ""
    __DEFAULT_SPLIT_BY__ = "EchoNumbers"

    def __init__(self, volumes: Union[MedicalVolume, Sequence[MedicalVolume]]):
        self.volumes = volumes

        # TODO: Remove series number as an attribute.
        # It should be directly accessed through the reference header.
        # This involves some backward compatible fixes.
        self.series_number = None
        self._from_file_args = {}

        self.temp_path = os.path.join(
            env.temp_dir(), self.NAME, strftime("%Y-%m-%d-%H-%M-%S", localtime())
        )
        self.tissues = []
        self._metadata = {}

    def __validate_scan__(self) -> bool:
        """Validate this scan (usually done by checking dicom header tags, if available).
        Returns:
            bool: `True` if scan metadata is valid, `False` otherwise.
        """
        return True

    def get_metadata(self, key: Any, default=None):
        """Get metadata for the scan.

        The metadata is either stored in the ``self._metadata`` property
        if it is not a part of the standard DICOM header or the reference
        DICOM does not exist.

        Args:
            key (Any): The attribute to fetch.
            default (Any, optional): The default value to return.
                Set to ``False`` to raise an error when metadata is not found in
                either the scan metadata or DICOM headers (if available).
                Defaults to ``None``.

        Returns:
            Any: The value.

        Raises:
            ValueError: If ``default == False`` and metadata not found
        """
        metadata = self._metadata.get(key, None)
        if metadata is None and self.ref_dicom is not None:
            metadata = self.ref_dicom[key].value if key in self.ref_dicom else None
        if metadata is None and default is False:
            raise KeyError(f"Metadata '{key}' not found")
        elif metadata is None:
            return default
        else:
            return metadata

    def get_dimensions(self):
        """Get shape of volumes.

        All volumes in scan are assumed to be same dimension.

        Returns:
            tuple[int]: Shape of each volume in scan.
        """
        if isinstance(self.volumes, MedicalVolume):
            return self.volumes.shape
        return self.volumes[0].shape

    @property
    def ref_dicom(self):
        """The reference dicom.

        The reference dicom is defined as the first dicom of the first volume if
        ``self.volumes`` is a sequence of volumes.
        """
        vol = self.volumes[0] if isinstance(self.volumes, Sequence) else self.volumes
        headers = vol.headers(flatten=True)
        return headers[0] if headers is not None else None

    def __add_tissue__(self, new_tissue: Tissue):
        """Add a tissue to the list of tissues associated with this scan.

        Args:
            new_tissue (Tissue): Tissue to add.

        Raises:
            ValueError: If tissue already exists in list.
                For example, we cannot add FemoralCartilage twice to the list of tissues.
        """
        contains_tissue = any([tissue.ID == new_tissue.ID for tissue in self.tissues])
        if contains_tissue:
            raise ValueError("Tissue already exists")

        self.tissues.append(new_tissue)

    def to(self, device):
        """Moves volumes of this scan onto the appropriate device.

        This is an inplace operation. A new ScanSequence will not
        be returned.

        Args:
            device: The device to move to.

        Returns:
            self
        """
        if isinstance(self.volumes, MedicalVolume):
            self.volumes = self.volumes.to(device)
            return

        self.volumes = [v.to(device) for v in self.volumes]

        return self


class NonTargetSequence(ScanSequence):
    """
    Abstract class for scans that cannot serve as targets and have
    to be registered to some other target scan.

    Examples: Cubequant, Cones
    """

    @abstractmethod
    def interregister(self, target_path: str, mask_path: str = None):
        """Register this scan to the target scan - save as parameter in scan (volumes, subvolumes, etc).

        We use the term *interregister* to refer to registration between volumes of different scans.
        Conversely, *intraregister* refers to registering volumes from the same scan.

        If there are many subvolumes to interregister with the base scan,
        typically the following actions are used to reduce error accumulation
        with multiple registrations:

        1. Pick the volume with the highest SNR. Call this the base moving image.
        2. Register the base moving image to the target image using elastix.
        3. Capture the transformation file(s) detailing the transforms to go from
           base moving image -> target image.
        4. Apply these transformation(s) to the remaining volumes.

        Args:
            target_path (str): Path to NIfTI file storing scan.
                This scan will serve as the base for registration.
                Note the best target scan will typically have a high SNR.

            mask_path (str): Path to mask to use to use as focus points for registration.
        """
        pass  # pragma: no cover

    def __load_interregistered_files__(self, interregistered_dirpath: str):
        """Load the NIfTI files of the interregistered subvolumes.

        These subvolumes have already been registered to some target scan
        using the `interregister` function.

        Args:
            interregistered_dirpath (str): Directory path where interregistered volumes are stored.

        Returns:
            Tuple[Dict[int, MedicalVolume], List[float]]:
                A dictionary mapping echo time index -> MedicalVolume

        Raises:
            ValueError: If files are not of the name `<INTEGER>.nii.gz`
                (e.g. `0.nii.gz`, `000.nii.gz`, etc.) or if no interregistered files
                found in interregistered_dirpath.
        """
        _logger.info("Loading interregistered files")
        if "interregistered" not in interregistered_dirpath:
            raise ValueError("Invalid path for loading {} interregistered files".format(self.NAME))

        subfiles = os.listdir(interregistered_dirpath)
        subfiles = natsorted(subfiles)

        if len(subfiles) == 0:
            raise ValueError("No interregistered files found")

        indices = []
        subvolumes = []
        nifti_reader = NiftiReader()
        for subfile in subfiles:
            subfile_nums = re.findall(r"[-+]?\d*\.\d+|\d+", subfile)
            if len(subfile_nums) == 0:
                raise ValueError("{} is not an interregistered '.gz.nii' file.".format(subfile))

            subfile_num = int(subfile_nums[0])
            indices.append(subfile_num)

            filepath = os.path.join(interregistered_dirpath, subfile)
            subvolume = nifti_reader.load(filepath)

            subvolumes.append(subvolume)

        assert len(indices) == len(subvolumes), "Number of subvolumes mismatch"

        if len(subvolumes) == 0:
            raise ValueError("No interregistered files found")

        subvolumes_dict = {}
        for i in range(len(indices)):
            subvolumes_dict[indices[i]] = subvolumes[i]

        return subvolumes_dict

    def __dilate_mask__(
        self,
        mask_path: str,
        temp_path: str,
        dil_rate: float = preferences.mask_dilation_rate,
        dil_threshold: float = preferences.mask_dilation_threshold,
    ):
        """Dilate mask using gaussian blur and write to disk to use with Elastix.

        Args:
            mask_path (str | MedicalVolume): File path for mask or mask to use to use as
                focus points for registration. Mask must be binary.
            temp_path (str): Directory path to store temporary data.
            dil_rate (`float`, optional): Dilation rate (sigma).
                Defaults to ``preferences.mask_dilation_rate``.
            dil_threshold (`float`, optional): Threshold to binarize dilated mask.
                Must be between [0, 1]. Defaults to ``preferences.mask_dilation_threshold``.

        Returns:
            str: File path of dilated mask.

        Raises:
            FileNotFoundError: If `mask_path` not valid file.
            ValueError: If `dil_threshold` not in range [0, 1].
        """

        if dil_threshold < 0 or dil_threshold > 1:
            raise ValueError("'dil_threshold' must be in range [0, 1]")

        if isinstance(mask_path, MedicalVolume):
            mask = mask_path
        elif os.path.isfile(mask_path):
            mask = fio_utils.generic_load(mask_path, expected_num_volumes=1)
        else:
            raise FileNotFoundError("File {} not found".format(mask_path))

        dilated_mask = (
            sni.gaussian_filter(np.asarray(mask.volume, dtype=np.float32), sigma=dil_rate)
            > dil_threshold
        )
        fixed_mask = np.asarray(dilated_mask, dtype=np.int8)
        fixed_mask_filepath = os.path.join(io_utils.mkdirs(temp_path), "dilated-mask.nii.gz")

        dilated_mask_volume = MedicalVolume(fixed_mask, affine=mask.affine)
        dilated_mask_volume.save_volume(fixed_mask_filepath)

        return fixed_mask_filepath



"""Basic file constants to be shared with program"""

import os

# Elastix files

__DIR__ = os.path.abspath(os.path.dirname(__file__))
__OUT_DIR__ = os.path.abspath(os.path.expanduser("~/.dosma"))
_DOSMA_ELASTIX_FOLDER = os.path.join(__DIR__, "resources/elastix")
__PATH_TO_ELASTIX_FOLDER__ = os.path.join(__DIR__, "resources/elastix/params")

ELASTIX_AFFINE_PARAMS_FILE = os.path.join(__PATH_TO_ELASTIX_FOLDER__, "parameters-affine.txt")
ELASTIX_BSPLINE_PARAMS_FILE = os.path.join(__PATH_TO_ELASTIX_FOLDER__, "parameters-bspline.txt")
ELASTIX_RIGID_PARAMS_FILE = os.path.join(__PATH_TO_ELASTIX_FOLDER__, "parameters-rigid.txt")

ELASTIX_AFFINE_INTERREGISTER_PARAMS_FILE = os.path.join(
    __PATH_TO_ELASTIX_FOLDER__, "parameters-affine-interregister.txt"
)
ELASTIX_RIGID_INTERREGISTER_PARAMS_FILE = os.path.join(
    __PATH_TO_ELASTIX_FOLDER__, "parameters-rigid-interregister.txt"
)

MAPSS_ELASTIX_AFFINE_INTERREGISTER_PARAMS_FILE = os.path.join(
    __PATH_TO_ELASTIX_FOLDER__, "parameters-affine-interregister.txt"
)
MAPSS_ELASTIX_RIGID_INTERREGISTER_PARAMS_FILE = os.path.join(
    __PATH_TO_ELASTIX_FOLDER__, "parameters-rigid-interregister.txt"
)



"""Quantitative Double Echo in Steady State (qDESS) sequence."""
import logging
import math
import warnings
from copy import deepcopy
from typing import Sequence, Tuple

import numpy as np
import pydicom
from pydicom.tag import Tag

from dosma.core.med_volume import MedicalVolume
from dosma.core.quant_vals import T2
from dosma.models.seg_model import SegModel
from dosma.scan_sequences.scans import ScanSequence
from dosma.tissues.tissue import Tissue
from dosma.utils.cmd_line_utils import ActionWrapper

__all__ = ["QDess"]

_logger = logging.getLogger(__name__)


class QDess(ScanSequence):
    """qDESS MRI sequence.

    Quantitative double echo in steady state (qDESS) is a high-resolution scan that consists
    of two echos (S1, S2) that has shown high efficacy for analytic :math:`T_2` mapping.
    Because of its high resolution, qDESS has been shown to be a good candidate for automatic
    segmentation.

    DOSMA supports both automatic segmentation and analytical T2 solving for qDESS scans.
    Automated segmentation uses pre-trained convolutional neural networks (CNNs).

    References:
        B Sveinsson, AS Chaudhari, GE Gold, BA Hargreaves. A simple analytic method
        for estimating :math:`T_2` in the knee from DESS. Magnetic Resonance in Medicine,
        38:63-70 (2017). `[link] <https://www.ncbi.nlm.nih.gov/pubmed/28017730>`_
    """

    NAME = "qdess"

    # DESS DICOM header keys
    __GL_AREA_TAG__ = Tag(0x001910B6)
    __TG_TAG__ = Tag(0x001910B7)

    # DESS constants
    __NUM_ECHOS__ = 2
    __VOLUME_DIMENSIONS__ = 3

    def __init__(self, volumes: Sequence[MedicalVolume]):
        if len(volumes) != 2:
            raise ValueError("QDess currently only supports 2 volumes.")
        super().__init__(volumes)

    def __validate_scan__(self) -> bool:
        """Validate that the dicoms are of qDESS sequence.

        Returns:
            bool: `True` if has 2 echos, `False` otherwise.
        """
        return len(self.volumes) == self.__NUM_ECHOS__

    def segment(self, model: SegModel, tissue: Tissue, use_rss: bool = False):
        """Segment tissue in scan.

        Args:
            model (SegModel): Model to use for segmenting scans.
            tissue (Tissue): The tissue to segment.
            use_rss (`bool`, optional): If ``True``, use root-sum-of-squares (RSS) of
                echos for segmentation (preferred for built-in methods).
                If ``False``, use first echo for segmentation.

        Returns:
            MedicalVolume: Binary mask for segmented region.
        """
        tissue_names = (
            ", ".join([t.FULL_NAME for t in tissue])
            if isinstance(tissue, Sequence)
            else tissue.FULL_NAME
        )
        _logger.info(f"Segmenting {tissue_names}...")

        if use_rss:
            segmentation_volume = self.calc_rss()
        else:
            # Use first echo for segmentation.
            segmentation_volume = self.volumes[0]

        # Segment tissue and add it to list.
        mask = model.generate_mask(segmentation_volume)
        if isinstance(mask, dict):
            if not isinstance(tissue, Sequence):
                tissue = [tissue]
            for abbreviation, tis in zip([t.STR_ID for t in tissue], tissue):
                tis.set_mask(mask[abbreviation])
                self.__add_tissue__(tis)
        else:
            assert isinstance(tissue, Tissue)
            tissue.set_mask(mask)
            self.__add_tissue__(tissue)

        return mask

    def generate_t2_map(
        self,
        tissue: Tissue = None,
        suppress_fat: bool = False,
        suppress_fluid: bool = False,
        beta: float = 1.2,
        gl_area: float = None,
        tg: float = None,
        tr: float = None,
        te: float = None,
        alpha: float = None,
        diffusivity: float = 1.25e-9,
        t1: float = None,
        nan_bounds: Tuple[float, float] = (0, 100),
        nan_to_num: float = 0.0,
        decimals: int = 1,
    ):
        """Generate 3D T2 map.

        Spoiler amplitude (``gl_area``) and duration (``tg``) must be specified if dicom header
        does not contain relevant private tags. If dicom header is unavailable, ``tr``, ``te``,
        and ``alpha`` must also be specified.

        All array-like arguments must be the same dimensions as the echo 1 and echo 2 volumes.

        Args:
            tissue (Tissue, optional): Tissue to generate T2 map for.
                If provided, the resulting quantitative value map will
                be added to the list of quantitative values for the tissue.
            suppress_fat (`bool`, optional): Suppress fat region in T2 computation.
                This can help reduce noise.
            suppress_fluid (`bool`, optional): Suppress fluid region in T2 computation.
                Fluid-nulled image is calculated as ``S1 - beta*S2``.
            beta (float, optional): Beta value used for suppressing fluid.
            gl_area (float, optional): Spoiler area.
                Defaults to value in dicom private tag '0x001910b6'.
                Required if dicom header unavailable or private tag missing.
            tg (float, optional): Spoiler duration (in microseconds).
                Defaults to value in dicom private tag ``0x001910b7``.
                Required if dicom header unavailable or private tag missing.
            tr (float, optional): Repitition time (in milliseconds).
                Required if dicom header unavailable.
            te (float, optional): Echo time (in milliseconds).
                Required if dicom header unavailable.
            alpha (float or array-like): Flip angle in degrees.
                Required if dicom header unavailable.
            diffusivity (float or array-like): Estimated diffusivity.
            t1 (float or array-like): Estimated t1 in milliseconds.
                Defaults to ``tissue.T1_EXPECTED`` if ``tissue`` provided.
            nan_bounds (float or Tuple[float, float], optional): The closed interval
                (``[a, b]``) for valid t2 values. All values outside of this interval will
                be set to ``nan``.
            nan_to_num (float): Value to be used to fill NaN values. If ``None``, values
                will not be replaced.
            decimals (int): Number of decimal places to round to. If ``None``, values
                will not be rounded.

        Returns:
            qv.T2: T2 fit map.
        """

        if self.volumes is None:
            raise ValueError("volumes and ref_dicom fields must be initialized")

        if (
            self.get_metadata(self.__GL_AREA_TAG__, gl_area) is None
            or self.get_metadata(self.__TG_TAG__, tg) is None
        ):
            raise ValueError(
                "Dicom headers do not contain tags for `gl_area` and `tg`. Please input manually"
            )

        xp = self.volumes[0].device.xp
        ref_dicom = self.ref_dicom if self.ref_dicom is not None else pydicom.Dataset()

        r, c, num_slices = self.volumes[0].volume.shape
        subvolumes = self.volumes

        # Split echos
        echo_1 = subvolumes[0].volume
        echo_2 = subvolumes[1].volume

        # All timing in seconds
        TR = (float(ref_dicom.RepetitionTime) if tr is None else tr) * 1e-3
        TE = (float(ref_dicom.EchoTime) if te is None else te) * 1e-3
        Tg = (float(ref_dicom[self.__TG_TAG__].value) if tg is None else tg) * 1e-6
        T1 = (float(tissue.T1_EXPECTED) if t1 is None else t1) * 1e-3

        # Flip Angle (degree -> radians)
        alpha = float(ref_dicom.FlipAngle) if alpha is None else alpha
        alpha = math.radians(alpha)
        if np.allclose(math.sin(alpha / 2), 0):
            warnings.warn("sin(flip angle) is close to 0 - t2 map may fail.")

        GlArea = float(ref_dicom[self.__GL_AREA_TAG__].value) if gl_area is None else gl_area

        Gl = GlArea / (Tg * 1e6) * 100
        gamma = 4258 * 2 * math.pi  # Gamma, Rad / (G * s).
        dkL = gamma * Gl * Tg

        # Simply math
        k = (
            xp.power((xp.sin(alpha / 2)), 2)
            * (1 + xp.exp(-TR / T1 - TR * xp.power(dkL, 2) * diffusivity))
            / (1 - xp.cos(alpha) * xp.exp(-TR / T1 - TR * xp.power(dkL, 2) * diffusivity))
        )

        c1 = (TR - Tg / 3) * (xp.power(dkL, 2)) * diffusivity

        # T2 fit
        mask = xp.ones([r, c, num_slices])

        ratio = mask * echo_2 / echo_1
        ratio = xp.nan_to_num(ratio)

        # have to divide division into steps to avoid overflow error
        t2map = -2000 * (TR - TE) / (xp.log(abs(ratio) / k) + c1)

        t2map = xp.nan_to_num(t2map)

        # Filter calculated T2 values that are below 0ms and over 100ms
        if nan_bounds is not None:
            lower, upper = nan_bounds
            t2map[(t2map < lower) | (t2map > upper)] = xp.nan
        if nan_to_num is not None:
            t2map = (
                xp.nan_to_num(t2map)
                if isinstance(nan_to_num, bool)
                else xp.nan_to_num(t2map, nan=nan_to_num)
            )

        if decimals is not None:
            t2map = xp.around(t2map, decimals)

        if suppress_fat:
            t2map = t2map * (echo_1 > 0.15 * xp.max(echo_1))

        if suppress_fluid:
            vol_null_fluid = echo_1 - beta * echo_2
            t2map = t2map * (vol_null_fluid > 0.1 * xp.max(vol_null_fluid))

        t2_map_wrapped = subvolumes[0]._partial_clone(volume=t2map, headers=True)
        t2_map_wrapped = T2(t2_map_wrapped)

        if tissue is not None:
            tissue.add_quantitative_value(t2_map_wrapped)

        return t2_map_wrapped

    def calc_rss(self):
        """Calculate root-sum-of-squares (RSS) of two echos.

        Returns:
            MedicalVolume: Volume corresponding to RSS of two echos.
        """
        return self._combine_echoes("rss")

    def _combine_echoes(self, method="rss"):
        """Combine two echoes.

        Args:
            method (str, optional): Supported combination methods are:
                * ``"rss"`` (default): The root-sum-of-squares of two echoes.
                * ``"rms"``: The root-mean-square of two echoes.

        Returns:
            MedicalVolume: Volume with RMS of two echos.
        """
        xp = self.volumes[0].device.xp

        if self.volumes is None:
            raise ValueError("Volumes must be initialized")

        assert len(self.volumes) == 2, "2 Echos expected"

        echo1 = xp.asarray(self.volumes[0].volume, dtype=xp.float64)
        echo2 = xp.asarray(self.volumes[1].volume, dtype=xp.float64)

        assert (~xp.iscomplex(echo1)).all() and (~xp.iscomplex(echo2)).all()

        if method == "rss":
            vol = xp.sqrt(echo1 ** 2 + echo2 ** 2)
        elif method == "rms":
            vol = xp.sqrt((echo1 ** 2 + echo2 ** 2) / 2)
        else:
            raise ValueError(f"`method={method}` is not supported")

        mv = deepcopy(self.volumes[0])
        mv.volume = vol

        return mv

    def _save(self, metadata, save_dir, fname_fmt=None, **kwargs):
        default_fmt = {MedicalVolume: "echo-{}"}
        default_fmt.update(fname_fmt if fname_fmt else {})
        return super()._save(metadata, save_dir, fname_fmt=default_fmt, **kwargs)

    @classmethod
    def cmd_line_actions(cls):
        """
        Provide command line information (such as name, help strings, etc)
        as list of dictionary.
        """

        segment_action = ActionWrapper(
            name=cls.segment.__name__,
            help="generate automatic segmentation",
            param_help={"use_rss": "use root sum of squares (RSS) of two echos for segmentation"},
            alternative_param_names={"use_rss": ["rss"]},
        )
        generate_t2_map_action = ActionWrapper(
            name=cls.generate_t2_map.__name__,
            aliases=["t2"],
            param_help={
                "suppress_fat": "suppress computation on low SNR fat regions",
                "suppress_fluid": "suppress computation on fluid regions",
                "beta": "constant for calculating fluid-nulled image (S1-beta*S2)",
                "gl_area": "GL Area. Defaults to value in dicom tag '0x001910b6'",
                "tg": "Gradient time (in microseconds). "
                "Defaults to value in dicom tag '0x001910b7'.",
                "alpha": "Flip angle in degrees. Defaults to value in dicom tag '0x00181314'.",
                "diffusivity": "Estimated diffusivity. Defaults to 1.25e-9",
            },
            help="generate T2 map",
        )

        return [(cls.segment, segment_action), (cls.generate_t2_map, generate_t2_map_action)]



"""Standardized orientation convention and utilities.

Medical image orientation convention is library and image format (DICOM, NIfTI, etc.)
dependent and is often difficult to interpret. This makes it challenging to intelligently
and rapidly reformat images.

We adopt a human readable orientation representation
for the dimensions and define utilities to convert between different orientation formats
from current libraries (Nibabel, PyDicom, ITK, etc).

Orientations are represented by string axis codes:

- ``"LR"``: left to right; ``"RL"``: right to left
- ``"PA"``: posterior to anterior; ``"AP"``: anterior to posterior
- ``"IS"``: inferior to superior; ``"SI"``: superior to inferior

A :class:`MedicalVolume` object with orientation ``("SI", "AP", "LR")`` has an
array where the first dimension spans superior -> inferior, the second dimension
spans anterior -> posterior, and the third dimension spans left -> right. Voxel
at (i,j,k) index ``(0,0,0)`` would be the (superior, anterior, left) corner.

In many cases, images are not acquired in the standard plane convention, but rather
in a rotated frame. In this case, the orientations correspond to the closest axis
the a particular dimension.

Two general conventions are followed:

- All orientations are in patient voxel coordinates. Image data from (i, j, k)
  corresponds to the voxel at array position ``arr[i,j,k]``.
- Left: corresponds to patient (not observer) left,
  right: corresponds to patient (not observer) right.

We adopt the RAS+ standard (as defined by NIfTI) for orienting our images.
The ``+`` in RAS+ indicates that all directions point to the increasing direction.
i.e. from -x to x:.

Image spacing, direction, and global origin are represented by a 4x4 affine matrix (:math:`A`) and
is identical to the nibabel affine matrix
(see `nibabel <https://nipy.org/nibabel/coordinate_systems.html>`_).
The affine matrix converts pixel coordinates (i, j, k) into world (NIfTI) coordinates (x, y, z).

.. math::

    \\begin{bmatrix} x\\\\y\\\\z\\\\1\\end{bmatrix} = A
    \\begin{bmatrix} i\\\\j\\\\k\\\\1\\end{bmatrix}


For example,

.. math::

    \\begin{bmatrix} x\\\\y\\\\z\\\\1 \\end{bmatrix} =
    \\begin{bmatrix} 0 & 0 & 1.5 & -61.6697\\\\-0.3125 & 0 & 0 & 50.8516\\\\
    0 & -0.3125 & 0 & 88.5876\\\\0 & 0 & 0 & 1 \\end{bmatrix}
    \\begin{bmatrix} i\\\\j\\\\k\\\\1\\end{bmatrix}

For details on how the affine matrix is used for reformatting see
:class:`dosma.core.MedicalVolume`.

"""
from typing import Sequence, Union

import nibabel.orientations as nibo
import numpy as np

__all__ = [
    "to_affine",
    "get_transpose_inds",
    "get_flip_inds",
    "orientation_nib_to_standard",
    "orientation_standard_to_nib",
    "SAGITTAL",
    "CORONAL",
    "AXIAL",
]


SAGITTAL = ("SI", "AP", "LR")
CORONAL = ("SI", "LR", "AP")
AXIAL = ("AP", "LR", "SI")

__EXPECTED_ORIENTATION_TUPLE_LEN__ = 3
__SUPPORTED_ORIENTATIONS__ = ["LR", "RL", "PA", "AP", "IS", "SI"]
__ORIENTATIONS_TO_AXIS_ID__ = {"LR": 0, "RL": 0, "PA": 1, "AP": 1, "IS": 2, "SI": 2}


def __check_orientation__(orientation: tuple):
    """Check if orientation tuple defines a valid orientation.

    Args:
        orientation (tuple[str]): Image orientation in standard orientation format.

    Raises:
        ValueError: If orientation tuple is invalid.
    """
    is_orientation_format = (
        len(orientation) == __EXPECTED_ORIENTATION_TUPLE_LEN__
        and sum([type(o) is str for o in orientation]) == __EXPECTED_ORIENTATION_TUPLE_LEN__
    )

    orientation_str_exists = (
        sum([o in __SUPPORTED_ORIENTATIONS__ for o in orientation])
        == __EXPECTED_ORIENTATION_TUPLE_LEN__
    )

    orientation_ids = [__ORIENTATIONS_TO_AXIS_ID__[o] for o in orientation]
    unique_ids = len(orientation_ids) == len(set(orientation_ids))

    if not is_orientation_format or not orientation_str_exists or not unique_ids:
        raise ValueError(
            "Orientation format mismatch: Orientations must be tuple of strings of "
            "length {}".format(__EXPECTED_ORIENTATION_TUPLE_LEN__)
        )


def get_transpose_inds(curr_orientation: tuple, new_orientation: tuple):
    """Get indices for reordering planes from ``curr_orientation`` to ``new_orientation``.

    Only permuted order of reformatting the image planes is returned.
    For example, ``("SI", "AP", "LR")`` and ``("IS", "PA", "RL")`` will have no permuted
    indices because "SI"/"IS", "AP"/"PA" and "RL"/"LR" each correspond to the same
    plane.

    Args:
        curr_orientation (tuple[str]): Current image orientation.
        new_orientation (tuple[str]): New image orientation.

    Returns:
        tuple[int]: Axes to transpose to change orientation.

    Examples:
        >>> get_transpose_inds(("SI", "AP", "LR"), ("AP", "SI", "LR"))
        (1,0,2)
        >>> get_transpose_inds(("SI", "AP", "LR"), ("IS", "PA", "RL"))
        (0,1,2)
    """
    __check_orientation__(curr_orientation)
    __check_orientation__(new_orientation)

    curr_orientation_ids = [__ORIENTATIONS_TO_AXIS_ID__[o] for o in curr_orientation]
    new_orientation_ids = [__ORIENTATIONS_TO_AXIS_ID__[o] for o in new_orientation]

    if set(curr_orientation_ids) != set(new_orientation_ids):
        raise ValueError(
            "Orientation mismatch: Both curr_orientation and new_orientation "
            "must contain the same axes"
        )

    transpose_inds = [curr_orientation_ids.index(n_o) for n_o in new_orientation_ids]

    return tuple(transpose_inds)


def get_flip_inds(curr_orientation: tuple, new_orientation: tuple):
    """Get indices to flip from ``curr_orientation`` to ``new_orientation``.

    Args:
        curr_orientation (tuple[str]): Current image orientation.
        new_orientation (tuple[str]): New image orientation.

    Returns:
        list[int]: Axes to flip.

    Raises:
        ValueError: If mismatch in orientation indices. To avoid this error,
            use :func:`get_transpose_inds` prior to flipping.

    Examples:
        >>> get_transpose_inds(("SI", "AP", "LR"), ("IS", "AP", "RL"))
        [0, 2]
    """
    __check_orientation__(curr_orientation)
    __check_orientation__(new_orientation)

    curr_orientation_ids = [__ORIENTATIONS_TO_AXIS_ID__[o] for o in curr_orientation]
    new_orientation_ids = [__ORIENTATIONS_TO_AXIS_ID__[o] for o in new_orientation]

    if curr_orientation_ids != new_orientation_ids:
        raise ValueError(
            "All axis orientations (S/I, L/R, A/P) must be ordered. "
            "Use `get_transpose_inds` to reorder axes."
        )

    flip_axs_inds = []
    for i in range(__EXPECTED_ORIENTATION_TUPLE_LEN__):
        if curr_orientation[i] != new_orientation[i]:
            flip_axs_inds.append(i)

    return flip_axs_inds


# Nibabel to standard orientation conversion utils.
__nib_to_standard_orientation_map__ = {
    "R": "LR",
    "L": "RL",
    "A": "PA",
    "P": "AP",
    "S": "IS",
    "I": "SI",
}


def orientation_nib_to_standard(nib_orientation):
    """Convert Nibabel orientation to the standard dosma orientation format.

    Args:
        nib_orientation: a RAS+ tuple orientation used by Nibabel.

    Returns:
        tuple[str]: Image orientation in the standard orientation format.

    Examples:
        >>> orientation_nib_to_standard(("R", "A", "S"))
        ("LR", "PA", "IS")
    """
    orientation = []
    for symb in nib_orientation:
        orientation.append(__nib_to_standard_orientation_map__[symb])
    return tuple(orientation)


def orientation_standard_to_nib(orientation):
    """Convert standard dosma orientation format to Nibabel orientation.

    Args:
        orientation: Image orientation in the standard orientation format.

    Returns:
        tuple[str]: RAS+ tuple orientation used by Nibabel.

    Examples:
        >>> orientation_nib_to_standard(("LR", "PA", "IS"))
        ("R", "A", "S")
    """
    nib_orientation = []
    for symb in orientation:
        nib_orientation.append(symb[1])
    return tuple(nib_orientation)


def to_affine(
    orientation,
    spacing: Sequence[Union[int, float]] = None,
    origin: Sequence[Union[int, float]] = None,
):
    """Convert orientation, spacing, and origin data into affine matrix.

    Args:
        orientation (Sequence[str]): Image orientation in the standard orientation format
            (e.g. ``("LR", "AP", "SI")``).
        spacing (int(s) | float(s)): Number(s) corresponding to pixel spacing of each direction.
            If a single value, same pixel spacing is used for all directions.
            If sequence is less than length of ``orientation``, remaining direction have unit
            spacing (i.e. ``1``). Defaults to unit spacing ``(1, 1, 1)``
        origin (int(s) | float(s)): The ``(x0, y0, z0)`` origin for the scan.
            If a single value, same origin is used for all directions.
            If sequence is less than length of ``orientation``, remaining direction have standard
            origin (i.e. ``0``). Defaults to ``(0, 0, 0)``

    Returns:
        ndarray: A 4x4 ndarray representing the affine matrix.

    Examples:
        >>> to_affine(("SI", "AP", "RL"), spacing=(0.5, 0.5, 1.5), origin=(10, 20, 0))
        array([[-0. , -0. , -1.5,  10. ],
               [-0. , -0.5, -0. ,  20. ],
               [-0.5, -0. , -0. ,  30. ],
               [ 0. ,  0. ,  0. ,   1. ]])

    Note:
        This method assumes all direction follow the standard principal directions in the normative
        patient orientation. Moving along one direction of the array only moves along one fo the
        normative directions.
    """

    def _format_numbers(input, default_val, name, expected_num):
        """Formats (sequence of) numbers (spacing, origin) into standard 3-length tuple."""
        if input is None:
            return (default_val,) * expected_num
        if isinstance(input, (int, float)):
            return (input,) * expected_num

        if not isinstance(input, (np.ndarray, Sequence)) or len(input) > expected_num:
            raise ValueError(
                f"`{name}` must be a real number or sequence (length<={expected_num}) "
                f"of real numbers. Got {input}"
            )
        input = tuple(input)

        if len(input) < expected_num:
            input += (default_val,) * (expected_num - len(input))
        assert len(input) == expected_num

        return input

    if len(orientation) == 2:
        orientation = _infer_orientation(orientation)
    __check_orientation__(orientation)
    spacing = _format_numbers(spacing, 1, "spacing", len(orientation))
    origin = _format_numbers(origin, 0, "origin", len(orientation))

    affine = np.eye(4)
    start_ornt = nibo.io_orientation(affine)
    end_ornt = nibo.axcodes2ornt(orientation_standard_to_nib(orientation))
    ornt = nibo.ornt_transform(start_ornt, end_ornt)

    transpose_idxs = ornt[:, 0].astype(np.int)
    flip_idxs = ornt[:, 1]

    affine[:3] = affine[:3][transpose_idxs]
    affine[:3] *= flip_idxs[..., np.newaxis]
    affine[:3, :3] *= np.asarray(spacing)
    affine[:3, 3] = np.asarray(origin)

    return affine


def _infer_orientation(orientation):
    """Infer 3-length orientation from 2-length orientation.

    Args:
        orientation: The incomplete orientation.

    Returns:
        tuple[str, str, str]: Standard orientation.
    """
    idxs = {__ORIENTATIONS_TO_AXIS_ID__[k] for k in orientation}
    if len(orientation) != 2 or len(idxs) != 2:
        raise ValueError(
            "`orientation` must be an incomplete orientation that encodes orthogonal directions"
        )

    missing_ornt = [k for k, v in __ORIENTATIONS_TO_AXIS_ID__.items() if v not in idxs][0]
    return tuple(orientation) + (missing_ornt,)



from dosma.models import oaiunet2d, stanford_qdess, util
from dosma.models.oaiunet2d import *  # noqa
from dosma.models.stanford_qdess import *  # noqa: F401, F403
from dosma.models.util import *  # noqa

__all__ = []
__all__.extend(util.__all__)
__all__.extend(oaiunet2d.__all__)
__all__.extend(stanford_qdess.__all__)



"""Models trained on the Osteoarthritis Initiative (OAI) dataset(s)."""

import os
from copy import deepcopy

import numpy as np

try:
    from keras.layers import BatchNormalization as BN
    from keras.layers import Concatenate, Conv2D, Conv2DTranspose, Dropout, Input, MaxPooling2D
    from keras.models import Model

    _SUPPORTS_KERAS = True
except ImportError:  # pragma: no-cover
    _SUPPORTS_KERAS = False  # pragma: no-cover

from dosma.core.med_volume import MedicalVolume
from dosma.core.orientation import SAGITTAL
from dosma.models.seg_model import KerasSegModel, whiten_volume

__all__ = ["OAIUnet2D", "IWOAIOAIUnet2D", "IWOAIOAIUnet2DNormalized"]


class OAIUnet2D(KerasSegModel):
    """Model trained in Chaudhari et al. IWOAI 2018

    Original Github: https://github.com/akshaysc/msk_segmentation
    """

    ALIASES = ["oai-unet2d", "oai_unet2d"]

    sigmoid_threshold = 0.5

    def __load_keras_model__(self, input_shape, force_weights=False):
        """Generate Unet 2D model

        Args:
            input_shape: tuple of input size - format: (height, width, 1)

        Returns:
            A Keras model

        Raises:
            ValueError: If ``input_size`` is not tuple or dimensions
                or ``input_size`` does not match (height, width, 1)
        """
        if not _SUPPORTS_KERAS:
            raise ImportError(
                "`oaiunet2d` segmentation models depend on tensorflow/keras backends. "
                "Install them with `pip install tensorflow; pip install keras`"
            )

        if type(input_shape) is not tuple or len(input_shape) != 3 or input_shape[2] != 1:
            raise ValueError("input_size must be a tuple of size (height, width, 1)")

        nfeatures = [2 ** feat * 32 for feat in np.arange(6)]
        depth = len(nfeatures)

        conv_ptr = []

        # input layer
        inputs = Input(input_shape)

        # step down convolutional layers
        pool = inputs
        for depth_cnt in range(depth):

            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(pool)
            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(conv)

            conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
            conv = Dropout(rate=0.0)(conv)

            conv_ptr.append(conv)

            # Only maxpool till penultimate depth
            if depth_cnt < depth - 1:

                # If size of input is odd, only do a 3x3 max pool
                xres = conv.shape.as_list()[1]
                if xres % 2 == 0:
                    pooling_size = (2, 2)
                elif xres % 2 == 1:
                    pooling_size = (3, 3)

                pool = MaxPooling2D(pool_size=pooling_size)(conv)

        # step up convolutional layers
        for depth_cnt in range(depth - 2, -1, -1):

            deconv_shape = conv_ptr[depth_cnt].shape.as_list()
            deconv_shape[0] = None

            # If size of input is odd, then do a 3x3 deconv
            if deconv_shape[1] % 2 == 0:
                unpooling_size = (2, 2)
            elif deconv_shape[1] % 2 == 1:
                unpooling_size = (3, 3)

            up = Concatenate(axis=3)(
                [
                    Conv2DTranspose(
                        nfeatures[depth_cnt], (3, 3), padding="same", strides=unpooling_size
                    )(conv),
                    conv_ptr[depth_cnt],
                ]
            )

            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(up)
            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(conv)

            conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
            conv = Dropout(rate=0.00)(conv)

        # combine features
        recon = Conv2D(1, (1, 1), padding="same", activation="sigmoid")(conv)

        model = Model(inputs=[inputs], outputs=[recon])

        return model

    def generate_mask(self, volume: MedicalVolume):
        vol_copy = deepcopy(volume)

        # reorient to the sagittal plane
        vol_copy.reformat(SAGITTAL, inplace=True)

        vol = vol_copy.volume
        vol = self.__preprocess_volume__(vol)

        # reshape volumes to be (slice, x, y, 1)
        v = np.transpose(vol, (2, 0, 1))
        v = np.expand_dims(v, axis=-1)
        model = self.seg_model

        mask = model.predict(v, batch_size=self.batch_size, verbose=1)
        mask = (mask > self.sigmoid_threshold).astype(np.uint8)

        # reshape mask to be (x, y, slice)
        mask = np.transpose(np.squeeze(mask, axis=-1), (1, 2, 0))

        vol_copy.volume = mask

        # reorient to match with original volume
        vol_copy.reformat(volume.orientation, inplace=True)

        return vol_copy

    def __preprocess_volume__(self, volume: np.ndarray):
        # TODO: Remove epsilon if difference in performance difference is not large.
        return whiten_volume(volume, eps=1e-8)


class IWOAIOAIUnet2D(OAIUnet2D):
    """
    Model trained by Team 6 in the 2019 IWOAI Segmentation Challenge.

    References:
        Desai, et al., "The International Workshop on Osteoarthritis Imaging Knee
        MRI Segmentation Challenge: A Multi-Institute Evaluation and Analysis
        Framework on a Standardized Dataset." arXiv preprint arXiv:2004.14003
        (2020). `[link] <https://arxiv.org/abs/2004.14003>`_
    """

    ALIASES = ["iwoai-2019-t6"]
    _WEIGHTS_FILE = "iwoai-2019-unet2d_fc-tc-pc-men_weights.h5"

    def __init__(self, input_shape, weights_path, force_weights=False):
        if not force_weights and os.path.basename(weights_path) != self._WEIGHTS_FILE:
            raise ValueError(f"Weights {weights_path} not supported for {type(self)}")
        super().__init__(input_shape, weights_path)

    def __load_keras_model__(self, input_shape):
        if type(input_shape) is not tuple or len(input_shape) != 3 or input_shape[2] != 1:
            raise ValueError("input_size must be a tuple of size (height, width, 1)")

        nfeatures = [2 ** feat * 32 for feat in np.arange(6)]
        depth = len(nfeatures)

        conv_ptr = []

        # input layer
        inputs = Input(input_shape)

        # step down convolutional layers
        pool = inputs
        for depth_cnt in range(depth):

            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(pool)
            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(conv)

            conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
            conv = Dropout(rate=0.0)(conv)

            conv_ptr.append(conv)

            # Only maxpool till penultimate depth
            if depth_cnt < depth - 1:

                # If size of input is odd, only do a 3x3 max pool
                xres = conv.shape.as_list()[1]
                if xres % 2 == 0:
                    pooling_size = (2, 2)
                elif xres % 2 == 1:
                    pooling_size = (3, 3)

                pool = MaxPooling2D(pool_size=pooling_size)(conv)

        # step up convolutional layers
        for depth_cnt in range(depth - 2, -1, -1):

            deconv_shape = conv_ptr[depth_cnt].shape.as_list()
            deconv_shape[0] = None

            # If size of input is odd, then do a 3x3 deconv
            if deconv_shape[1] % 2 == 0:
                unpooling_size = (2, 2)
            elif deconv_shape[1] % 2 == 1:
                unpooling_size = (3, 3)

            up = Concatenate(axis=3)(
                [
                    Conv2DTranspose(
                        nfeatures[depth_cnt], (3, 3), padding="same", strides=unpooling_size
                    )(conv),
                    conv_ptr[depth_cnt],
                ]
            )

            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(up)
            conv = Conv2D(
                nfeatures[depth_cnt],
                (3, 3),
                padding="same",
                activation="relu",
                kernel_initializer="he_normal",
            )(conv)

            conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
            conv = Dropout(rate=0.00)(conv)

        # combine features
        recon = Conv2D(4, (1, 1), padding="same", activation="sigmoid")(conv)

        model = Model(inputs=[inputs], outputs=[recon])

        return model

    def generate_mask(self, volume: MedicalVolume):
        vol_copy = deepcopy(volume)

        # reorient to the sagittal plane
        vol_copy.reformat(SAGITTAL, inplace=True)

        vol = vol_copy.volume
        vol = self.__preprocess_volume__(vol)

        # reshape volumes to be (slice, x, y, 1)
        v = np.transpose(vol, (2, 0, 1))
        v = np.expand_dims(v, axis=-1)
        model = self.seg_model

        mask = model.predict(v, batch_size=self.batch_size, verbose=1)
        mask = (mask > self.sigmoid_threshold).astype(np.uint8)

        # reshape mask to be (x, y, slice, classes)
        mask = np.transpose(mask, (1, 2, 0, 3))

        vols = {}
        for i, category in enumerate(["fc", "tc", "pc", "men"]):
            vol_cp = deepcopy(vol_copy)
            vol_cp.volume = mask[..., i]

            # reorient to match with original volume
            vol_cp.reformat(volume.orientation, inplace=True)
            vols[category] = vol_cp

        return vols

    def __preprocess_volume__(self, volume: np.ndarray):
        return volume


class IWOAIOAIUnet2DNormalized(IWOAIOAIUnet2D):
    """
    Extension of model trained by Team 6 in the 2019 IWOAI Segmentation
    Challenge (with normalization).

    This model uses the same architecture as :class:`IWOAIOAIUnet2D`,
    but pre-processes the input data by zero-mean, unit-std normalization.

    References:
        Desai, et al., "The International Workshop on Osteoarthritis Imaging Knee
        MRI Segmentation Challenge: A Multi-Institute Evaluation and Analysis
        Framework on a Standardized Dataset." arXiv preprint arXiv:2004.14003
        (2020).
    """

    ALIASES = ("iwoai-2019-t6-normalized",)
    _WEIGHTS_FILE = "iwoai-2019-unet2d-normalized_fc-tc-pc-men_weights.h5"

    def __preprocess_volume__(self, volume: np.ndarray):
        return whiten_volume(volume)



"""Initialize and parse command line arguments for DOSMA.

This module is the entry point for executing DOSMA from the command line.
The DOSMA library is critical for processing.

To use this file, it must be run as a module from the parent directory::

    $ python -m dosma/cli ...

Examples:
    Run T2 fitting on Subject 01, Series 007, a quantitative DESS (qDESS) scan,
    for femoral cartilage::

    python -m dosma/cli --dicom subject01/dicoms/007/ --save subject01/data/ \
        qdess --fc generate_t2_map

Hint:
    Run ``python -m dosma/cli --help`` for a detailed description of different
    command line arguments.
"""

import argparse
import ast
import functools
import inspect
import logging
import os
import time
import warnings
from collections import defaultdict
from typing import Sequence

from dosma.core.io.format_io import ImageDataFormat
from dosma.core.quant_vals import QuantitativeValueType as QV
from dosma.defaults import preferences
from dosma.models.seg_model import SegModel
from dosma.models.util import SUPPORTED_MODELS, get_model, model_from_config
from dosma.msk import knee
from dosma.scan_sequences.mri.cones import Cones
from dosma.scan_sequences.mri.cube_quant import CubeQuant
from dosma.scan_sequences.mri.mapss import Mapss
from dosma.scan_sequences.mri.qdess import QDess
from dosma.scan_sequences.scans import ScanSequence
from dosma.tissues.tissue import Tissue
from dosma.utils import env
from dosma.utils.logger import setup_logger

SUPPORTED_QUANTITATIVE_VALUES = [QV.T2, QV.T1_RHO, QV.T2_STAR]

DEBUG_KEY = "debug"

DICOM_KEY = "dicom"
SAVE_KEY = "save"
LOAD_KEY = "load"
IGNORE_EXT_KEY = "ignore_ext"
SPLIT_BY_KEY = "split_by"

GPU_KEY = "gpu"
NUM_WORKERS_KEY = "num-workers"

SCAN_KEY = "scan"
SCAN_ACTION_KEY = "scan_action"

SEGMENTATION_MODEL_KEY = "model"
SEGMENTATION_CONFIG_KEY = "config"
SEGMENTATION_WEIGHTS_DIR_KEY = "weights_dir"
SEGMENTATION_BATCH_SIZE_KEY = "batch_size"

TISSUES_KEY = "tissues"

SUPPORTED_SCAN_TYPES = [Cones, CubeQuant, Mapss, QDess]
BASIC_TYPES = [bool, str, float, int, list, tuple]

_logger = logging.getLogger(__name__)


class CommandLineScanContainer:
    def __init__(
        self,
        scan_type: type,
        dicom_path,
        load_path,
        ignore_ext: bool = False,
        group_by=None,
        num_workers=0,
        **kwargs,
    ):
        """The class for command-line handling around :class:`ScanSequence`.

        The command line interface for :class:`ScanSequence` data is highly structured,
        particularly in data saving and loading to support easy and accurate command-line
        compatibility.

        This class overloads some standard functionality in :class:`ScanSequence`
        (:func:`save`, :func:`load`). When methods are not implemented, this class provides
        access directly to attributes/methods of the underlying scan instantiation.
        For example, if ``scan_type=QDess``, the following will call
        :func:`QDess.generate_t2_map`:

        >>> cli_scan = CommandLineScanContainer(QDess, dicom_path="/path/to/qdess/scan")
        >>> cli_scan.generate_t2_map(...)  # this calls cli_scan.scan.generate_t2_map

        Data is loaded either from the ``dicom_path`` or the ``load_path``. If both are specified,
        the data is loaded from the ``dicom_path``.

        Args:
            scan_type (type): A scan type. Should be subclass of `ScanSequence`.
            dicom_path (str): The dicom path. This value can be ``None``, but must
                be explicitly set.
            load_path (str): The load path. This value can be ``None``, but must be
                explicitly set.
            ignore_ext (bool, optional): If ``True``, ignore extensions when loading
                dicom data. See :func:`DicomReader.load` for details.
            group_by (optional): The value(s) to group dicoms by. See :func:`DicomReader.load`
                for details.
            num_workers (int, optional): Number of works for loading scan.

        Attributes:
            scan_type (type): The scan type to instantiate.
            scan (ScanSequence): The instantiated scan.
            generic_args (Dict[str, Any]): Generic duck typed parameter names and values.
                If parameters with this name are part of the method signature, they will
                automatically be set to the values in this dictionary. Keys include:
                    * "num_workers": Number of cpu workers to use.
                    * "max_workers": Alias for "num_workers" in some functions
                    * "verbose": Verbosity
                    * "show_pbar": Show progress bar.

        Raises:
            NotADirectoryError: If ``dicom_path`` is not a path to a directory.
        """
        self.scan_type = scan_type

        if (dicom_path is not None) and (not os.path.isdir(dicom_path)):
            if load_path is not None:
                warnings.warn(
                    "Dicom_path {} not found. Will load data from {}".format(dicom_path, load_path)
                )
            else:
                raise NotADirectoryError("{} is not a directory".format(dicom_path))

        # Only use dicoms if the path exists and path contains files.
        is_dicom_available = (dicom_path is not None) and (os.path.isdir(dicom_path))

        # If dicom_path is specified and exists, assume user wants to start from scratch with the
        # dicoms. load_path is ignored.
        group_by = group_by if group_by is not None else scan_type.__DEFAULT_SPLIT_BY__
        if is_dicom_available:
            scan = scan_type.from_dicom(
                dicom_path, group_by=group_by, ignore_ext=ignore_ext, num_workers=num_workers
            )
        else:
            scan = self.load(load_path, num_workers=num_workers)

        self.scan = scan
        self.generic_args = {
            "num_workers": num_workers,
            "max_workers": num_workers,
            "verbose": True,
            "show_pbar": True,
        }

    def __getattr__(self, name):
        attr = getattr(self.scan, name)
        if callable(attr):
            params = inspect.signature(attr).parameters
            params = params.keys() & self.generic_args.keys()
            kwargs = {k: self.generic_args[k] for k in params}
            if len(kwargs):
                attr = functools.partial(attr, **kwargs)
        return attr

    def load(self, path: str, num_workers: int = 0):
        """Command line interface loading scan data.

        ``self.scan_type`` must be set before calling this function.

        Args
            path (str): Path to pickle file or directory where data is stored.
            num_workers (int, optional): Number of workers to use to load data.

        Returns:
            ScanSequence: Scan of type ``self.scan_type``.

        Raises:
            ValueError: If path to load data from cannot be determined.

        Examples:
            >>> cli_scan.load("/path/to/pickle/file")  # load data from pickle file
            >>> cli_scan.load("/path/to/directory")  # load data from directory
        """
        scan_type = self.scan_type

        file_path = None
        if os.path.isfile(path):
            file_path = path
        elif os.path.isdir(path) and scan_type.NAME:
            fname = f"{scan_type.NAME}.data"
            _paths = (
                os.path.join(path, fname),
                os.path.join(self._save_dir(path, create_dir=False), fname),
            )
            for _path in _paths:
                if os.path.isfile(_path):
                    file_path = _path
                    break
        if file_path is None:
            raise ValueError(f"Cannot load {scan_type.__name__} data from path '{path}'")

        return scan_type.load(file_path, num_workers)

    def _save_dir(self, dir_path: str, create_dir: bool = True):
        """Returns directory path specific to this scan.

        Formatted as '`base_load_dirpath`/`scan.NAME`'.

        Args:
            dir_path (str): Directory path where all data is stored.
            create_dir (`bool`, optional): If `True`, creates directory if it doesn't exist.

        Returns:
            str: Data directory path for this scan.
        """
        scan_type = self.scan_type
        folder_id = scan_type.NAME

        name_len = len(folder_id) + 2  # buffer
        if scan_type.NAME not in dir_path[-name_len:]:
            scan_dirpath = os.path.join(dir_path, folder_id)
        else:
            scan_dirpath = dir_path

        # scan_dirpath = os.path.join(scan_dirpath, folder_id)

        if create_dir:
            os.makedirs(scan_dirpath, exist_ok=True)

        return scan_dirpath

    def save(
        self,
        path: str,
        save_custom: bool = True,
        image_data_format: ImageDataFormat = None,
        num_workers: int = 0,
    ):
        path = self._save_dir(path, create_dir=True)
        return self.scan.save(path, save_custom, image_data_format, num_workers)


def get_nargs_for_basic_type(base_type: type):
    if base_type in [str, float, int]:
        return 1
    elif base_type in [list, tuple]:
        return "+"


def add_tissues(parser: argparse.ArgumentParser):
    for tissue in knee.SUPPORTED_TISSUES:
        parser.add_argument(
            "--%s" % tissue.STR_ID,
            action="store_const",
            default=False,
            const=True,
            help="analyze %s" % tissue.FULL_NAME,
        )


def parse_tissues(vargin: dict):
    tissues = []
    for tissue in knee.SUPPORTED_TISSUES:
        t = tissue()
        if (
            t.STR_ID in vargin.keys()
            and vargin[t.STR_ID]
            and t.STR_ID not in [x.STR_ID for x in tissues]
        ):
            load_path = vargin[LOAD_KEY]
            if load_path:
                t.load_data(load_path)

            tissues.append(t)

    # if no tissues are specified, do computation for all supported tissues
    if len(tissues) == 0:
        _logger.info("No tissues specified, computing for all supported tissues...")
        tissues = []
        for tissue in knee.SUPPORTED_TISSUES:
            t = tissue()
            if t.STR_ID not in [x.STR_ID for x in tissues]:
                load_path = vargin[LOAD_KEY]
                if load_path:
                    t.load_data(load_path)

                tissues.append(t)

    analysis_str = "Tissue(s): "
    for tissue in tissues:
        analysis_str += "%s, " % tissue.FULL_NAME

    _logger.info(analysis_str)

    return tissues


def add_segmentation_subparser(parser):
    parser.add_argument(
        "--%s" % SEGMENTATION_WEIGHTS_DIR_KEY,
        type=str,
        nargs=1,
        required=True,
        help="path to directory with weights",
    )
    parser.add_argument(
        "--%s" % SEGMENTATION_MODEL_KEY,
        choices=SUPPORTED_MODELS,
        nargs="?",
        default=None,
        help="built-in model to use for segmentation. Choices: %s" % SUPPORTED_MODELS,
    )
    parser.add_argument(
        "--%s" % SEGMENTATION_CONFIG_KEY,
        type=str,
        default=None,
        help="config file for non-built-in model",
    )
    parser.add_argument(
        "--%s" % SEGMENTATION_BATCH_SIZE_KEY,
        metavar="B",
        type=int,
        default=preferences.segmentation_batch_size,
        nargs="?",
        help="batch size for inference. Default: %d" % preferences.segmentation_batch_size,
    )

    return parser


def handle_segmentation(vargin, scan: ScanSequence, tissue: Tissue):
    if not vargin[SEGMENTATION_MODEL_KEY] and not vargin[SEGMENTATION_CONFIG_KEY]:
        raise ValueError(
            "Either `--{}` or `--{}` must be specified".format(
                SEGMENTATION_MODEL_KEY, SEGMENTATION_CONFIG_KEY
            )
        )

    segment_weights_path = vargin[SEGMENTATION_WEIGHTS_DIR_KEY][0]
    if isinstance(tissue, Sequence):
        weights = [t.find_weights(segment_weights_path) for t in tissue]
        assert all(weights_file == weights[0] for weights_file in weights)
        weights_path = weights[0]
    else:
        weights_path = tissue.find_weights(segment_weights_path)

    # Load model
    dims = scan.get_dimensions()
    # TODO: Input shape should be determined by combination of model + scan.
    # Currently fixed in 2D plane
    input_shape = (dims[0], dims[1], 1)
    if vargin[SEGMENTATION_MODEL_KEY]:
        # Use built-in model
        model = get_model(
            vargin[SEGMENTATION_MODEL_KEY], input_shape=input_shape, weights_path=weights_path
        )
    else:
        # Use config
        model = model_from_config(
            vargin[SEGMENTATION_CONFIG_KEY],
            weights_dir=segment_weights_path,
            input_shape=input_shape,
        )
    model.batch_size = vargin[SEGMENTATION_BATCH_SIZE_KEY]

    return model


CUSTOM_TYPE_TO_HANDLE_DICT = {SegModel: handle_segmentation}


def add_custom_argument(parser, param_type):
    # handle all custom arguments except tissues
    has_custom_argument = False
    if param_type is SegModel:
        add_segmentation_subparser(parser)
        has_custom_argument = True

    return has_custom_argument


def add_base_argument(
    parser: argparse.ArgumentParser,
    param_name,
    param_type,
    param_default,
    param_help,
    additional_param_names: list = None,
):
    if additional_param_names is None:
        additional_param_names = []

    # TODO: Clean up this code block to properly do syntax parsing.
    try:
        if param_type not in BASIC_TYPES:
            param_type = extract_basic_type(param_type)
    except (AttributeError, TypeError):
        raise TypeError(
            "Parameter '{}' - type '{}' not in BASIC_TYPES".format(param_name, param_type)
        )

    # add default value to param help
    has_default = param_default is not inspect._empty
    if has_default:
        param_help = "%s. Default: %s" % (param_help, param_default)

    if additional_param_names:
        param_names = ["--%s" % n for n in additional_param_names]
    else:
        param_names = []

    param_names.append("--%s" % param_name)

    if param_type is bool:
        if not has_default:
            raise ValueError("All boolean parameters must have a default value.")

        parser.add_argument(
            *param_names,
            action="store_%s" % (str(not param_default).lower()),
            dest=param_name,
            help=param_help,
        )
        return

    # all other values with default have this parameter
    nargs_no_default = get_nargs_for_basic_type(param_type)
    nargs = "?" if has_default else nargs_no_default

    parser.add_argument(
        *param_names,
        nargs=nargs,
        default=param_default if has_default else None,
        dest=param_name,
        help=param_help,
        required=not has_default,
    )


def parse_basic_type(val, param_type):
    if param_type not in BASIC_TYPES:
        param_type = extract_basic_type(param_type)

    if type(val) is param_type:
        return val

    if param_type in [list, tuple]:
        return param_type(val)

    nargs = get_nargs_for_basic_type(param_type)
    if type(val) is list and nargs == 1:
        return val[0]
    return param_type(val) if val else val


def extract_basic_type(param_type):
    """Extracts basic types from ``typing`` aliases.

    Args:
        param_type (typing._GenericAlias): A generic alias
            (e.g. ``typing.Tuple``, ``typing.List``).

    Returns:
        type: The basic type.
    """
    try:
        # Python 3.5 / 3.6
        return param_type.__extra__
    except AttributeError:
        # Python 3.7/3.8/3.9
        return param_type.__origin__


def add_scans(dosma_subparser):
    for scan in SUPPORTED_SCAN_TYPES:
        supported_actions = scan.cmd_line_actions()

        # skip scans that are not supported on the command line
        if len(supported_actions) == 0:
            continue
        scan_name = scan.NAME
        scan_parser = dosma_subparser.add_parser(scan.NAME, help="analyze %s sequence" % scan_name)
        add_tissues(scan_parser)

        if not len(supported_actions):
            raise ValueError("No supported actions for scan %s" % scan_name)

        scan_subparser = scan_parser.add_subparsers(
            description="%s subcommands" % scan.NAME, dest=SCAN_ACTION_KEY
        )

        for action, action_wrapper in supported_actions:
            func_signature = inspect.signature(action)
            func_name = action_wrapper.name
            aliases = action_wrapper.aliases
            action_parser = scan_subparser.add_parser(
                func_name, aliases=aliases, help=action_wrapper.help
            )

            parameters = func_signature.parameters
            for param_name in parameters.keys():
                param = parameters[param_name]
                param_type = param.annotation
                param_default = param.default

                if param_name == "self" or param_type is Tissue:
                    continue

                param_help = action_wrapper.get_param_help(param_name)
                alternative_param_names = action_wrapper.get_alternative_param_names(param_name)

                if param_type is inspect._empty:
                    raise ValueError(
                        "scan %s, action %s, param %s does not have an annotation. "
                        "Use pytying in the method declaration" % (scan.NAME, func_name, param_name)
                    )

                # see if the type is a custom type, if not handle it as a basic type
                is_custom_arg = add_custom_argument(action_parser, param_type)
                if is_custom_arg:
                    continue

                add_base_argument(
                    action_parser,
                    param_name,
                    param_type,
                    param_default,
                    param_help=param_help,
                    additional_param_names=alternative_param_names,
                )

        scan_parser.set_defaults(func=handle_scan)


def _find_tissue_groups(vargin, tissues: Sequence[Tissue]):
    """Finds groups of tissues that can be segmented together.

    Some segmentation models have multiple weight files for different tissues.
    Many models have one weight file for multiple tissues (i.e. multi-class segmentation).

    This function matches tissues with their corresponding weight file.
    If multiple tissues share a single weight file, they will map to the same weight file,
    allowing multiple tissues to be segmented simultaneously.

    This is a temporary fix for segmenting multiple classes.
    It should not be extended or used as precedence for future development.
    """
    if not isinstance(tissues, Sequence):
        assert isinstance(tissues, Tissue)
        tissues = [tissues]

    weights_dir = vargin[SEGMENTATION_WEIGHTS_DIR_KEY][0]
    weights_to_tissues = defaultdict(list)
    for tissue in tissues:
        weights_to_tissues[tissue.find_weights(weights_dir)].append(tissue)

    return weights_to_tissues


def _build_params(vargin, scan, parameters, tissue=None):
    param_dict = {}
    for param_name in parameters.keys():
        param = parameters[param_name]
        param_type = param.annotation

        if param_name == "self":
            continue

        if param_type is Tissue:
            assert tissue is not None
            param_dict["tissue"] = tissue
            continue

        if param_type in CUSTOM_TYPE_TO_HANDLE_DICT:
            param_dict[param_name] = CUSTOM_TYPE_TO_HANDLE_DICT[param_type](vargin, scan, tissue)
        else:
            param_dict[param_name] = parse_basic_type(vargin[param_name], param_type)
    return param_dict


def handle_scan(vargin):

    scan_name = vargin[SCAN_KEY]
    _logger.info("Analyzing {}...".format(scan_name))
    scan = None

    for p_scan in SUPPORTED_SCAN_TYPES:
        if p_scan.NAME == scan_name:
            scan = p_scan
            break

    scan = CommandLineScanContainer(
        scan,
        dicom_path=vargin[DICOM_KEY],
        load_path=vargin[LOAD_KEY],
        ignore_ext=vargin[IGNORE_EXT_KEY],
        split_by=vargin[SPLIT_BY_KEY] if vargin[SPLIT_BY_KEY] else scan.__DEFAULT_SPLIT_BY__,
        num_workers=vargin[NUM_WORKERS_KEY],
    )

    tissues = vargin["tissues"]
    scan_action = scan_action_str = vargin[SCAN_ACTION_KEY]

    p_action = None
    for action, action_wrapper in scan.cmd_line_actions():
        if scan_action == action_wrapper.name or scan_action in action_wrapper.aliases:
            p_action = action
            break

    # search for name in the cmd_line actions
    action = p_action

    if action is None:
        scan.save(vargin[SAVE_KEY], image_data_format=preferences.image_data_format)
        return

    func_signature = inspect.signature(action)
    parameters = func_signature.parameters
    if scan_action_str == "segment":
        weights_to_tissues = _find_tissue_groups(vargin, tissues)
        for _weights_file, seg_tissues in weights_to_tissues.items():
            if len(seg_tissues) == 1:
                seg_tissues = seg_tissues[0]
            param_dict = _build_params(vargin, scan, parameters, seg_tissues)
            getattr(scan, action.__name__)(**param_dict)
    else:
        if "tissue" in func_signature.parameters.keys():
            for tissue in tissues:
                param_dict = _build_params(vargin, scan, parameters, tissue)
                getattr(scan, action.__name__)(**param_dict)
        else:
            param_dict = _build_params(vargin, scan, parameters)
            getattr(scan, action.__name__)(**param_dict)

    scan.save(vargin[SAVE_KEY], image_data_format=preferences.image_data_format)
    for tissue in tissues:
        tissue.save_data(vargin[SAVE_KEY], data_format=preferences.image_data_format)

    return scan


def parse_dicom_tag_splitby(vargin_str):
    if not vargin_str:
        return vargin_str

    try:
        parsed_str = ast.literal_eval(vargin_str)
        return parsed_str
    except Exception:
        return vargin_str


def parse_args(f_input=None):
    """Parse arguments given through command line (argv)

    :raises ValueError if dicom path is not provided
    :raises NotADirectoryError if dicom path does not exist or is not a directory
    """
    parser = argparse.ArgumentParser(
        prog="DOSMA",
        description="A deep-learning powered open source MRI analysis pipeline",
        epilog="Either `--dicom` or `--load` must be specified. "
        "If both are given, `--dicom` will be used",
    )
    parser.add_argument("--%s" % DEBUG_KEY, action="store_true", help="use debug mode")

    # Dicom and results paths
    parser.add_argument(
        "--d",
        "--%s" % DICOM_KEY,
        metavar="D",
        type=str,
        default=None,
        nargs="?",
        dest=DICOM_KEY,
        help="path to directory storing dicom files",
    )
    parser.add_argument(
        "--l",
        "--%s" % LOAD_KEY,
        metavar="L",
        type=str,
        default=None,
        nargs="?",
        dest=LOAD_KEY,
        help="path to data directory to load from",
    )
    parser.add_argument(
        "--s",
        "--%s" % SAVE_KEY,
        metavar="S",
        type=str,
        default=None,
        nargs="?",
        dest=SAVE_KEY,
        help="path to data directory to save to. Default: L/D",
    )
    parser.add_argument(
        "--%s" % IGNORE_EXT_KEY,
        action="store_true",
        default=False,
        dest=IGNORE_EXT_KEY,
        help="ignore .dcm extension when loading dicoms. Default: False",
    )
    parser.add_argument(
        "--%s" % SPLIT_BY_KEY,
        metavar="G",
        type=str,
        default=None,
        nargs="?",
        dest=SPLIT_BY_KEY,
        help="override dicom tag to split volumes by (eg. `EchoNumbers`)",
    )

    parser.add_argument(
        "--%s" % GPU_KEY,
        metavar="G",
        type=str,
        default=None,
        nargs="?",
        dest=GPU_KEY,
        help="gpu id. Default: None",
    )

    parser.add_argument(
        "--%s" % NUM_WORKERS_KEY,
        metavar="G",
        type=int,
        default=0,
        dest=NUM_WORKERS_KEY,
        help="num cpu workers. Default: 0",
    )

    # Add preferences
    preferences_flags = preferences.cmd_line_flags()
    for flag in preferences_flags.keys():
        argparse_kwargs = preferences_flags[flag]
        argparse_kwargs["dest"] = flag
        aliases = argparse_kwargs.pop("aliases", None)
        name = argparse_kwargs.pop("name", None)  # noqa: F841
        parser.add_argument(*aliases, **argparse_kwargs)

    subparsers = parser.add_subparsers(help="sub-command help", dest=SCAN_KEY)
    add_scans(subparsers)

    # MSK knee parser
    knee.knee_parser(subparsers)

    start_time = time.time()
    if f_input:
        args = parser.parse_args(f_input)
    else:
        args = parser.parse_args()

        # Only initialize logger if called from command line.
        # If UI is using it, the logger should be initialized by the UI.
        setup_logger(env.log_file_path())

    vargin = vars(args)

    if vargin[DEBUG_KEY]:
        env.debug(True)

    gpu = vargin[GPU_KEY]

    _logger.debug(vargin)

    if gpu is not None:
        os.environ["CUDA_VISIBLE_DEVICES"] = gpu
    elif "CUDA_VISIBLE_DEVICES" not in os.environ:
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

    # parse and update preferences
    for flag in preferences_flags.keys():
        preferences.set(flag, vargin[flag])

    dicom_path = vargin[DICOM_KEY]
    load_path = vargin[LOAD_KEY]

    if not dicom_path and not load_path:
        raise ValueError("Must provide path to dicoms or path to load data from")

    save_path = vargin[SAVE_KEY]
    if not save_path:
        save_path = load_path if load_path else "%s/data" % dicom_path
        vargin[SAVE_KEY] = save_path

    if not os.path.isdir(save_path):
        os.makedirs(save_path)

    tissues = parse_tissues(vargin)
    vargin["tissues"] = tissues

    vargin[SPLIT_BY_KEY] = parse_dicom_tag_splitby(vargin[SPLIT_BY_KEY])

    args.func(vargin)

    time_elapsed = time.time() - start_time
    _logger.info("Time Elapsed: {:.2f} seconds".format(time.time() - start_time))

    return time_elapsed


if __name__ == "__main__":
    parse_args()



from __future__ import print_function

import numpy as np

import matplotlib

matplotlib.use("TkAgg")


class IndexTracker:
    def __init__(self, ax, x):
        self.ax = ax
        self.im = None
        self.ind = 0
        self.x = x

    def onscroll(self, event):
        if event.button == "down":
            self.ind = min(self.ind + 1, self.num_slices - 1)
        elif event.button == "up":
            self.ind = max(self.ind - 1, 0)

        self.update()

    def update(self):
        x_im = self._x_normalized
        x_im = np.squeeze(x_im[:, :, self.ind, :])
        if self.im is None:
            self.im = self.ax.imshow(x_im, cmap="gray")
            self.ax.get_xaxis().set_ticks([])
            self.ax.get_yaxis().set_ticks([])

        self.im.set_data(x_im)
        self.ax.set_ylabel("slice %s" % (self.ind + 1))
        self.im.axes.figure.canvas.draw()

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        assert type(value) is np.ndarray and (
            value.ndim == 3 or value.ndim == 4
        ), "image must be 3d (grayscale) or 4d (rgb) ndarray"
        if value.ndim == 3:
            value = value[..., np.newaxis]

        self._x = value
        self.num_slices = self._x.shape[2]
        self._x_normalized = self._x

        self.update()



import os
import unittest

import h5py
import numpy as np

from dosma.core.io.nifti_io import NiftiReader
from dosma.models.oaiunet2d import IWOAIOAIUnet2D, IWOAIOAIUnet2DNormalized
from dosma.models.seg_model import whiten_volume
from dosma.tissues.femoral_cartilage import FemoralCartilage

import keras.backend as K

from .. import util


@unittest.skipIf(not util.is_data_available(), "unittest data is not available")
class TestIWOAIOAIUnet2D(unittest.TestCase):
    def test_segmentation(self):
        """Check that segmentation works as expected from ported version."""
        classes = ["fc", "tc", "pc", "men"]
        expected_seg = np.load(
            os.path.join(
                util.UNITTEST_DATA_PATH, "datasets/oai/expected/test_001_V00-iwoai-2019-t6.npy"
            )
        )

        scan = NiftiReader().load(
            os.path.join(util.UNITTEST_DATA_PATH, "datasets/oai/test_001_V00.nii.gz")
        )

        tissue = FemoralCartilage()
        tissue.find_weights(os.path.join(os.path.dirname(__file__), "../../weights/iwoai-2019-t6"))
        dims = scan.volume.shape
        input_shape = (dims[0], dims[1], 1)
        model = IWOAIOAIUnet2D(input_shape=input_shape, weights_path=tissue.weights_file_path)
        masks = model.generate_mask(scan)
        K.clear_session()

        for i, tissue in enumerate(classes):
            assert np.all(masks[tissue].volume == expected_seg[..., i])


@unittest.skipIf(not util.is_data_available(), "unittest data is not available")
class TestIWOAIOAIUnet2DNormalized(unittest.TestCase):
    def test_h5_nifti_same(self):
        with h5py.File(
            os.path.join(util.UNITTEST_DATA_PATH, "datasets/oai/2000001_V00.h5"), "r"
        ) as f:
            h5_data = f["volume"][:]
        scan = NiftiReader().load(
            os.path.join(util.UNITTEST_DATA_PATH, "datasets/oai/test_001_V00.nii.gz")
        )
        assert np.all(scan.volume.astype(np.float32) == h5_data)

        # Issue #50 (https://github.com/ad12/DOSMA/issues/50)
        # h5_wh = whiten_volume(np.float32(h5_data))
        # scan_wh = whiten_volume(np.float32(scan.volume))
        # assert np.all(scan_wh == h5_wh)

    def test_normalized_h5df_data(self):
        """Compare zero-mean, unit std. dev. normalization with h5 scans.

        We are doing this comparison to ensure that the normalization done by DOSMA produces the
        expected normalzied output for data stored in h5df files specifically.

        Note that because of precision issues, normalizing arrays of different precision can
        result in different volumes.
        """
        with h5py.File(
            os.path.join(util.UNITTEST_DATA_PATH, "datasets/oai/2000001_V00_w.h5"), "r"
        ) as f:
            expected = f["volume"][:]

        with h5py.File(
            os.path.join(util.UNITTEST_DATA_PATH, "datasets/oai/2000001_V00.h5"), "r"
        ) as f:
            h5_data = f["volume"][:]
        normalized = whiten_volume(h5_data)

        assert np.all(normalized == expected)

    # def test_same_w(self):
    #     import h5py
    #     from dosma.models.seg_model import whiten_volume
    #     with h5py.File(os.path.join(util.UNITTEST_DATA_PATH, "datasets/oai/2000001_V00_w.h5"),
    #                    "r") as f:
    #         whitened = f["volume"][:]
    #
    #     with h5py.File(os.path.join(util.UNITTEST_DATA_PATH, "datasets/oai/2000001_V00.h5"), "r") as f:  # noqa: E501
    #         h5_data = f["volume"][:]
    #     h5_data = whiten_volume(h5_data)
    #
    #     assert np.all(whitened == h5_data)
    #
    # def test_same_whitened(self):
    #     import h5py
    #     from dosma.models.seg_model import whiten_volume
    #     with h5py.File(os.path.join(util.UNITTEST_DATA_PATH, "datasets/oai/2000001_V00_w.h5"), "r") as f:  # noqa: E501
    #         h5_data = f["volume"][:]
    #
    #     scan = NiftiReader().load(os.path.join(
    #         util.UNITTEST_DATA_PATH, "datasets/oai/test_001_V00.nii.gz"
    #     ))
    #
    #     scan._volume = whiten_volume(scan._volume.astype(np.float32))
    #     assert np.all(scan.volume == h5_data)

    def test_segmentation(self):
        classes = ["fc", "tc", "pc", "men"]
        expected_seg = np.load(
            os.path.join(
                util.UNITTEST_DATA_PATH,
                "datasets/oai/expected/test_001_V00-iwoai-2019-t6-normalized.npy",
            )
        )

        scan = NiftiReader().load(
            os.path.join(util.UNITTEST_DATA_PATH, "datasets/oai/test_001_V00.nii.gz")
        )

        tissue = FemoralCartilage()
        tissue.find_weights(
            os.path.join(os.path.dirname(__file__), "../../weights/iwoai-2019-t6-normalized")
        )
        dims = scan.volume.shape
        input_shape = (dims[0], dims[1], 1)
        model = IWOAIOAIUnet2DNormalized(
            input_shape=input_shape, weights_path=tissue.weights_file_path
        )
        masks = model.generate_mask(scan)
        K.clear_session()

        for i, tissue in enumerate(classes):
            pred = masks[tissue].volume.astype(np.bool)
            gt = expected_seg[..., i].astype(np.bool)
            dice = 2 * np.sum(pred & gt) / np.sum(pred.astype(np.uint8) + gt.astype(np.uint8))
            # Zero-mean normalization of 32-bit vs 64-bit data results in slightly different
            # estimations of the mean and standard deviation.
            # However, when both volumes are compared pre-normalization
            # using np.all() both volumes are the same (see :meth:`test_h5_nifti_same`).
            # This slight difference in the image can affect network performance.
            # As a placeholder, we assert that the dice between the expected and
            # produced segmentations for each tissue must be greater than 99%
            #
            # Update: We found for this particular model and example pair, the segmentations
            # achieve a dice score of 1.0 for all tissues.
            # We enforce that the predicted mask must be equal to the expected mask.
            assert dice >= 0.99, "{}: {:0.6f}".format(tissue, dice)
            assert np.all(
                masks[tissue].volume == expected_seg[..., i]
            ), f"Segmentation not same for {tissue}"



from abc import ABC, abstractmethod

import numpy as np

from dosma.core.med_volume import MedicalVolume
from dosma.defaults import preferences

try:
    import keras.backend as K
except ImportError:  # pragma: no cover
    pass


class SegModel(ABC):
    """
    Args:
        input_shape (Tuple[int] or List[Tuple[int]]): Shapes(s) for initializing input(s)
            into model in format (height, width, channels).
        weights_path (str): filepath to weights used to initialize Keras model
        force_weights (`bool`, optional): If `True`, load weights without any checking.
            Keras/Tensorflow only.
    """

    ALIASES = [""]  # each segmentation model must have an alias

    def __init__(self, input_shape, weights_path, force_weights=False):
        self.batch_size = preferences.segmentation_batch_size
        self.seg_model = self.build_model(input_shape, weights_path)

    @abstractmethod
    def build_model(self, input_shape, weights_path):
        """
        Builds a segmentation model architecture and loads weights.

        Args:
            input_shape: Input shape of volume
            weights_path:

        Returns:
            a segmentation model that can be used for segmenting tissues (a Keras/TF/PyTorch model)
        """
        pass

    @abstractmethod
    def generate_mask(self, volume: MedicalVolume):
        """Segment the MRI volumes.

        Args:
            volume (MedicalVolume): Volume to segment in proper orientation.

        Returns:
            :class:`MedicalVolume` or List[MedicalVolume]: Volumes are binarized (0,1)
                uint8 3D ndarray of shape ``volume.shape``.

        Raises:
            ValueError: If volumes is not 3D ndarray
                or if tissue is not a string or not in list permitted tissues.

        """
        pass

    def __call__(self, *args, **kwargs):
        return self.generate_mask(*args, **kwargs)

    def __preprocess_volume__(self, volume: np.ndarray):
        """
        Preprocess volume prior to putting as input into segmentation network
        :param volume: a numpy array
        :return: a preprocessed numpy array
        """
        return volume

    def __postprocess_volume__(self, volume: np.ndarray):
        """
        Post-process logits (probabilities) or binarized mask
        :param volume: a numpy array
        :return: a postprocessed numpy array
        """
        return volume


class KerasSegModel(SegModel):
    """
    Abstract wrapper for Keras model used for semantic segmentation
    """

    def build_model(self, input_shape, weights_path=None):
        keras_model = self.__load_keras_model__(input_shape)
        if weights_path:
            keras_model.load_weights(weights_path)

        return keras_model

    @abstractmethod
    def __load_keras_model__(self, input_shape):
        """
        Build Keras architecture

        :param input_shape: tuple or list of tuples for initializing input(s) into Keras model

        :return: a Keras model
        """
        pass

    def __del__(self):
        K.clear_session()


# ============================ Preprocessing utils ============================
__VOLUME_DIMENSIONS__ = 3
__EPSILON__ = 1e-8


def whiten_volume(x: np.ndarray, eps: float = 0.0):
    """Whiten volumes by mean and std of all pixels.

    Args:
        x (ndarray): 3D numpy array (MRI volumes)
        eps (float, optional): Epsilon to avoid division by 0.

    Returns:
        ndarray: A numpy array with mean ~ 0 and standard deviation ~ 1
    """
    if len(x.shape) != __VOLUME_DIMENSIONS__:
        raise ValueError(f"Input has {x.ndims} dimensions. Expected {__VOLUME_DIMENSIONS__}")

    return (x - np.mean(x)) / (np.std(x) + eps)



"""Analysis for tibial cartilage.

Attributes:
    BOUNDS (dict): Upper bounds for quantitative values.
"""
import os
import warnings

import numpy as np
import pandas as pd

from dosma.core.device import get_array_module
from dosma.core.med_volume import MedicalVolume
from dosma.core.quant_vals import QuantitativeValueType
from dosma.defaults import preferences
from dosma.tissues.tissue import Tissue, largest_cc
from dosma.utils import geometry_utils, io_utils

import matplotlib.pyplot as plt

# milliseconds
BOUNDS = {
    QuantitativeValueType.T2: 60.0,
    QuantitativeValueType.T1_RHO: 100.0,
    QuantitativeValueType.T2_STAR: 50.0,
}

__all__ = ["TibialCartilage"]


class TibialCartilage(Tissue):
    """Handles analysis and visualization for tibial cartilage."""

    ID = 4
    STR_ID = "tc"
    FULL_NAME = "tibial cartilage"

    # Expected quantitative values
    T1_EXPECTED = 1000  # milliseconds

    # Coronal Keys
    _ANTERIOR_KEY = 0
    _POSTERIOR_KEY = 1
    _CENTRAL_KEY = 2
    _CORONAL_KEYS = [_ANTERIOR_KEY, _CENTRAL_KEY, _POSTERIOR_KEY]

    # Saggital Keys
    _MEDIAL_KEY = 0
    _LATERAL_KEY = 1
    _SAGITTAL_KEYS = [_MEDIAL_KEY, _LATERAL_KEY]

    # Axial Keys
    _SUPERIOR_KEY = 0
    _INFERIOR_KEY = 1
    _TOTAL_AXIAL_KEY = -1

    def __init__(self, weights_dir=None, medial_to_lateral=None):
        super().__init__(weights_dir=weights_dir, medial_to_lateral=medial_to_lateral)

        self.regions_mask = None

    def unroll_axial(self, quant_map):
        mask = self.__mask__.volume

        assert (
            self.regions_mask is not None
        ), "region_mask not initialized. Should be initialized when mask is set"
        region_mask_sup_inf = self.regions_mask[..., 0]

        superior = (region_mask_sup_inf == self._SUPERIOR_KEY) * mask * quant_map
        superior[superior == 0] = np.nan
        superior = np.nanmean(superior, axis=0)

        inferior = (region_mask_sup_inf == self._INFERIOR_KEY) * mask * quant_map
        inferior[inferior == 0] = np.nan
        inferior = np.nanmean(inferior, axis=0)

        total = mask * quant_map
        total[total == 0] = np.nan
        total = np.nanmean(total, axis=0)

        return total, superior, inferior

    def split_regions(self, base_map):
        """Generate subregions for tibial cartilage.

        Tibial cartilage is split into subregions along the 3 major axes:
        superior/inferior (S/I), anterior/posterior (A/P), medial/lateral (M/L).
        M/L plateaus are computed with respect to the center of mass (COM)
        in the sagittal direction. A/C/P divisions are computed independently for each
        plateau using thirds of the distance between the minimum/maximum pixel in the A/P direction.
        S/I divisions are computed based on local COM for each 1D column with
        tissues in the S/I direction.

        The output is stored in `self.regions_mask`.

        Note:
            Previously, COM was used to split A/P subregions. However, COM is not as robust
            to split the data into 3 subregions. We use the method described in the reference
            below: A/C/P are split by thirds of the distance between the minimum/maximum pixel
            in the A/P direction. Note the minimum/maximum pixel method may not be a robust if
            there are a sufficient number of erroneous pixels. We do not sufficiently correct
            for erroneous pixels.

        Args:
            base_map (ndarray): Binary 3D mask with orientation (SI, AP, ML/LM).
                If `self.medial_to_lateral`, last dimension should be ML.

        References:
            Black, MS et al. Detecting Early Changes in ACL-Reconstructed Knees:
            Cluster Analysis of T2 Relaxation Times from 3 Months to 18 Months Post-Surgery.
            28th Annual Meeting of ISMRM, Sydney, Australia 2020.
        """
        xp = get_array_module(base_map)
        center_of_mass = geometry_utils.center_of_mass(base_map)  # zero indexed
        com_med_lat = int(xp.ceil(center_of_mass[2]))

        # M/L
        region_mask_med_lat = xp.zeros(base_map.shape)
        region_mask_med_lat[:, :, :com_med_lat] = (
            self._MEDIAL_KEY if self.medial_to_lateral else self._LATERAL_KEY
        )
        region_mask_med_lat[:, :, com_med_lat:] = (
            self._LATERAL_KEY if self.medial_to_lateral else self._MEDIAL_KEY
        )

        # S/I
        locs = base_map.sum(axis=0).nonzero()
        voxels = base_map[:, locs[0], locs[1]]
        com_sup_inf = xp.asarray(
            [
                int(xp.ceil(geometry_utils.center_of_mass(voxels[:, i])[0]))
                for i in range(voxels.shape[1])
            ]
        )
        region_mask_sup_inf = xp.full(base_map.shape, self._INFERIOR_KEY)
        for i in range(len(com_sup_inf)):
            region_mask_sup_inf[
                : com_sup_inf[i].item(), locs[0][i].item(), locs[1][i].item()
            ] = self._SUPERIOR_KEY

        # A/C/P
        region_mask_ant_post = xp.zeros(base_map.shape)
        for plateau in [slice(0, com_med_lat), slice(com_med_lat, None)]:
            cum_ap = xp.nonzero(base_map[..., plateau].sum(axis=(0, 2)))[0]
            min_ap = xp.min(cum_ap)
            ap_range = xp.max(cum_ap) - min_ap
            thresh1, thresh2 = (
                int(xp.ceil(min_ap + 1 / 3 * ap_range)),
                int(xp.ceil(min_ap + 2 / 3 * ap_range)),
            )
            region_mask_ant_post[:, :thresh1, plateau] = self._ANTERIOR_KEY
            region_mask_ant_post[:, thresh1:thresh2, plateau] = self._CENTRAL_KEY
            region_mask_ant_post[:, thresh2:, plateau] = self._POSTERIOR_KEY

        # # A/P
        # region_mask_ant_post = np.zeros(base_map.shape)
        # for plateau in [slice(0, com_med_lat), slice(com_med_lat, None)]:
        #     com = sni.measurements.center_of_mass(base_map[..., plateau])
        #     com_ant_post = int(np.ceil(com[1]))
        #     region_mask_ant_post[:, :com_ant_post, plateau] = self._ANTERIOR_KEY
        #     region_mask_ant_post[:, com_ant_post:, plateau] = self._POSTERIOR_KEY

        self.regions_mask = xp.stack(
            [region_mask_sup_inf, region_mask_ant_post, region_mask_med_lat], axis=-1
        )

    def __calc_quant_vals__(self, quant_map, map_type):
        subject_pid = self.pid

        super().__calc_quant_vals__(quant_map, map_type)

        assert (
            self.regions_mask is not None
        ), "region_mask not initialized. Should be initialized when mask is set"

        quant_map_volume = quant_map.volume
        mask = self.__mask__.volume

        quant_map_volume = mask * quant_map_volume

        axial_region_mask = self.regions_mask[..., 0]
        sagittal_region_mask = self.regions_mask[..., 1]
        coronal_region_mask = self.regions_mask[..., 2]

        axial_names = ["superior", "inferior", "total"]
        coronal_names = ["medial", "lateral"]
        sagittal_names = ["anterior", "posterior", "central"]

        pd_header = ["Subject", "Location", "Side", "Region", "Mean", "Std", "Median"]
        pd_list = []

        for axial in [self._SUPERIOR_KEY, self._INFERIOR_KEY, self._TOTAL_AXIAL_KEY]:
            if axial == self._TOTAL_AXIAL_KEY:
                axial_map = (axial_region_mask == self._SUPERIOR_KEY).astype(np.float32) + (
                    axial_region_mask == self._INFERIOR_KEY
                ).astype(np.float32)
                axial_map = axial_map.astype(np.bool)
            else:
                axial_map = axial_region_mask == axial

            for coronal in [self._MEDIAL_KEY, self._LATERAL_KEY]:
                for sagittal in [self._ANTERIOR_KEY, self._POSTERIOR_KEY, self._CENTRAL_KEY]:
                    curr_region_mask = (
                        quant_map_volume
                        * (coronal_region_mask == coronal)
                        * (sagittal_region_mask == sagittal)
                        * axial_map
                    )
                    curr_region_mask = curr_region_mask[curr_region_mask != 0]
                    # discard all values that are 0
                    c_mean = np.nanmean(curr_region_mask)
                    c_std = np.nanstd(curr_region_mask)
                    c_median = np.nanmedian(curr_region_mask)

                    row_info = [
                        subject_pid,
                        axial_names[axial],
                        coronal_names[coronal],
                        sagittal_names[sagittal],
                        c_mean,
                        c_std,
                        c_median,
                    ]

                    pd_list.append(row_info)

        # Generate 2D unrolled matrix
        total, superior, inferior = self.unroll_axial(quant_map.volume)

        df = pd.DataFrame(pd_list, columns=pd_header)
        qv_name = map_type.name
        maps = [
            {
                "title": "%s superior" % qv_name,
                "data": superior,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_superior" % qv_name,
                "raw_data_filename": "%s_superior.data" % qv_name,
            },
            {
                "title": "%s inferior" % qv_name,
                "data": inferior,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_inferior" % qv_name,
                "raw_data_filename": "%s_inferior.data" % qv_name,
            },
            {
                "title": "%s total" % qv_name,
                "data": total,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_total" % qv_name,
                "raw_data_filename": "%s_total.data" % qv_name,
            },
        ]

        self.__store_quant_vals__(maps, df, map_type)

    def set_mask(self, mask: MedicalVolume, use_largest_ccs=False):
        xp = get_array_module(mask.A)
        if use_largest_ccs:
            msk = xp.asarray(largest_cc(mask.A, num=2), dtype=xp.uint8)
        else:
            msk = xp.asarray(mask.A, dtype=xp.uint8)
        mask_copy = mask._partial_clone(volume=msk)
        super().set_mask(mask_copy)

        self.split_regions(self.__mask__.volume)

    def __save_quant_data__(self, dirpath):
        """Save quantitative data and 2D visualizations of tibial cartilage
        Check which quantitative values (T2, T1rho, etc) are defined for meniscus and analyze these
        1. Save 2D total, superficial, and deep visualization maps
        2. Save {'medial', 'lateral'}, {'anterior', 'posterior'},
            {'superior', 'inferior', 'total'} data to excel file

        :param dirpath: base filepath to save data
        """
        q_names = []
        dfs = []

        for quant_val in QuantitativeValueType:
            if quant_val.name not in self.quant_vals.keys():
                continue

            q_names.append(quant_val.name)
            q_val = self.quant_vals[quant_val.name]
            dfs.append(q_val[1])

            q_name_dirpath = io_utils.mkdirs(os.path.join(dirpath, quant_val.name.lower()))
            for q_map_data in q_val[0]:
                filepath = os.path.join(q_name_dirpath, q_map_data["filename"])
                xlabel = "Slice"
                ylabel = ""
                title = q_map_data["title"]
                data_map = q_map_data["data"]

                plt.clf()

                upper_bound = BOUNDS[quant_val]
                if preferences.visualization_use_vmax:
                    # Hard bounds - clipping
                    plt.imshow(data_map, cmap="jet", vmin=0.0, vmax=BOUNDS[quant_val])
                else:
                    # Try to use a soft bounds
                    if np.sum(data_map <= upper_bound) == 0:
                        plt.imshow(data_map, cmap="jet", vmin=0.0, vmax=BOUNDS[quant_val])
                    else:
                        warnings.warn(
                            "%s: Pixel value exceeded upper bound (%0.1f). Using normalized scale."
                            % (quant_val.name, upper_bound)
                        )
                        plt.imshow(data_map, cmap="jet")

                plt.xlabel(xlabel)
                plt.ylabel(ylabel)
                plt.title(title)
                clb = plt.colorbar()
                clb.ax.set_title("(ms)")
                plt.axis("tight")

                plt.savefig(filepath)

                # Save data
                raw_data_filepath = os.path.join(
                    q_name_dirpath, "raw_data", q_map_data["raw_data_filename"]
                )
                io_utils.save_pik(raw_data_filepath, data_map)

        if len(dfs) > 0:
            io_utils.save_tables(os.path.join(dirpath, "data.xlsx"), dfs, q_names)
